
import OpenGL.GL as gl

import os

from visvis import ssdf
from visvis.utils.pypoints import Pointset
#
from visvis.core.misc import basestring , getResourceDir

from visvis.text.text_base import AtlasTexture, FontManager
from visvis.text.text_base import correctVertices, simpleTextureDraw


class PrerenderedAtlas(AtlasTexture):
    pass


class Font:
    """ For the prerendered font system, the font consists of an atlas
    (one atlas per sans/serif/mono), and an info structure.
    """
    def __init__(self, info):
        self.info = info
        self.atlas = PrerenderedAtlas()
        self.atlas.SetData(info.data)


class PrerenderedFontManager(FontManager):
    
    def __init__(self):
        # load font data
        path = getResourceDir()
        self.s = ssdf.load(os.path.join(path, 'fonts.ssdf'))
        
        # list of fonts
        self.fonts = {}
    
    
    def GetFont(self, fontname):
        """ GetFont(fontname)
        
        Get a font instance. If that font was created earlier,
        that font is returned, otherwise it is created and stored
        for reuse.
        
        """
        if fontname in self.fonts:
            return self.fonts[fontname]
        elif hasattr(self.s, fontname):
            tmp = Font(self.s[fontname])
            self.fonts[fontname] = tmp
            return tmp
        else:
            raise ValueError("Invalid font name.")
    
    
    def Compile(self, textObject):
        """ Create a series of glyphs from the text in the textObject.
        From these Glyphs. Also the relative vertices are calculated,
        which are then corrected
        for angle and alignment in Position().
        """
        FontManager.Compile(self, textObject)
        
        # make invalid first
        textObject.Invalidate()
        
        # Get font object
        font = self.GetFont(textObject.fontName)
        
        # clear glyphs
        glyphs = []
        
        # Create reference character (used in Position)
        textObject._xglyph = Glyph(font, 'X', textObject.fontSize)
        
        # Get text string with escaped text converted to Unicode
        tt = self.ConvertEscapedText(textObject.text)
        
        
        # build list of glyphs, take sub/super scripting into account.
        escape = False
        styles = []
        style = None # Style to set
        for i in range(len(tt)):
            c = tt[i]
            if escape:
                g = Glyph(font, c, textObject.fontSize, styles)
                glyphs.append( g )
                escape = False
            elif c=='{':
                # Append style to the list
                if style:
                    styles.append(style)
                    style = None
            elif c=='}':
                # Remove style
                if styles:
                    styles.pop()
            elif c=='^':
                style = MiniStyle(2)
            elif c=='_':
                style = MiniStyle(1)
            elif c=='\x06':
                style = MiniStyle(0,False,True)
            elif c=='\x07':
                style = MiniStyle(0,True,False)
            elif c=='\\' and i+1<len(tt) and tt[i+1] in ['_^\x06\x07']:
                escape = True
            else:
                # create glyph (with new style (or not))
                g = Glyph(font, c, textObject.fontSize, styles+[style])
                glyphs.append( g )
                style = None
        
        # build arrays with vertices and coordinates
        x1, y1, z = 0, 0, 0
        vertices = Pointset(3)
        texCords = Pointset(2)
        for g in glyphs:
            x2 = x1 + g.sizex
            y2 = g.sizey
            #y2 = y1 - g.sizey
            dy = g.dy
            
            # append texture coordinates
            texCords.append(g.s1, g.t1)
            texCords.append(g.s2, g.t1)
            texCords.append(g.s2, g.t2)
            texCords.append(g.s1, g.t2)
            
            # set skewing for position
            skew = textObject.fontSize * g.skewFactor
            
            # append vertices
            vertices.append(x1+skew, y1+dy, z)
            vertices.append(x2+skew, y1+dy, z)
            vertices.append(x2, y2+dy, z)
            vertices.append(x1, y2+dy, z)
            
            # prepare for next glyph
            x1 = x1 + g.width + 1
        
        # Scale text according to global text size property
        fig = textObject.GetFigure()
        if fig:
            vertices *= fig._relativeFontSize
        
        # store calculations
        textObject._SetCompiledData(vertices, texCords)
    
    
    def Position(self, textObject):
        """ The name is ment as a verb. The vertices are corrected
        for angle and alignment.
        """
        FontManager.Position(self, textObject)
        
        # Get data
        vertices, texcoords = textObject._GetCompiledData()
        vertices = vertices.copy()
        
        # Use default algorithm to correct the vertices for alginment and angle
        correctVertices(textObject, vertices, textObject._xglyph.sizey)
                
        # Store
        textObject._SetFinalData(vertices, texcoords)
    
    
    def Draw(self, textObject, x=0, y=0, z=0):
        """ Draw the textobject.
        """
        FontManager.Draw(self, textObject)
        
        # Get data
        vertices, texCords = textObject._GetFinalData()
        
        # Translate
        if x or y or z:
            gl.glPushMatrix()
            gl.glTranslatef(x, y, z)
        
        # Draw
        atlas = self.GetFont(textObject.fontName).atlas
        simpleTextureDraw(vertices, texCords, atlas, textObject.textColor)
        
        # Un-translate
        if x or y or z:
            gl.glPopMatrix()


class Glyph(object):
    """ Glyph(font, char, size=12, styles=None)
    
    A glyph is a character. It is visualized by rendering
    the proper part from the texture stored in the Font object.
    
      * sizex and sizey represent the size of the glyph.
      * dy represents the offset in y direction (for sub/super scripts)
      * width specifies how much space there should be before the next char
      * s1 s2 t1 t2 represent texture coordinates
    
    """
    # the font.info contains
    # - a string of charcodes
    # - an array of origin 's
    # - an array of size's
    # - fontsize of the font in the data array
    
    def __init__(self, font, char, size=12, styles=None):
        
        # unwind the style for this glyph
        self.style = MiniStyle()
        if styles:
            for style in styles:
                self.style += style
        style = self.style
        
        # get info
        info = font.info
        
        # get asci code and check it
        if isinstance(char, basestring):
            ac = ord(char)
        elif isinstance(char, int):
            ac = char
        else:
            raise ValueError('To create a glyph, supply an int or character.')
        
        # do we have that char?
        if ac not in info.charcodes:  # ac < 32 or ac > 255:
            print("Warning: Cannot draw character %i! " % ord(char))
            ac = 32 # make space
        
        # default
        infoSize, infoOrigin, infoWidth = info.size, info.origin, info.width
        # should and can we display in italic or bold?
        # Note: italic is now realized by printing it skewed rather using the
        # italic glyphs. The reason is that when using the texture one would
        # see artifacts from neighbouring characters. Additionally, it's now
        # possible to mix bold and italic text, and one can make any supported
        # unicode character italic.
        #         if style.italic and ac in info.charcodes_i:
#             # italic text
#             infoSize, infoOrigin, infoWidth = (
#                 info.size_i, info.origin_i, info.width_i)
        if style.bold and ac in info.charcodes_b:
            # bold text
            infoSize, infoOrigin, infoWidth = (
                info.size_b, info.origin_b, info.width_b)
        
        # Find position in texture, normalized to texture coordinates
        x1 = infoOrigin[ac,0]
        x2 = x1 + infoSize[ac,0]
        tmp = float(info.data.shape[1])
        self.s1, self.s2 = (x1) / tmp, (x2-1) / tmp
        y1 = infoOrigin[ac,1]
        y2 = y1 + infoSize[ac,1]
        tmp = float(info.data.shape[0])
        self.t1, self.t2 = (y1) / tmp, (y2-1) / tmp
        
        # Define skew factor to handle italics correctly
        self.skewFactor = 0.0
        if style.italic:
            self.skewFactor = 0.5
        
        # calculate width on screen, given the size
        factor = size / float(info.fontsize)
        self.sizex = infoSize[ac,0] * factor
        self.sizey = infoSize[ac,1] * factor
        self.width = float(infoWidth[ac]) * factor # is spacing?
        
        smaller = 0.6
        self.dy = 0.0 # normal script
        if style.script == 1:
            # sub script
            self.dy = (1-smaller) * self.sizey
        if style.script:
            # super or subscript
            self.skewFactor *= smaller
            self.sizex = self.sizex * smaller
            self.sizey = self.sizey * smaller
            self.width = self.width * smaller  # - self.sizex * (1.0-smaller)


class MiniStyle:
    """ MiniStyle(script=0, bold=False, italic=False)
    
    Class that represents the style of characters (sub/super script,
    bold, and italic. Used when compiling the text.
    script = {0:'normal', 1:'sub', 2:'super'}
    
    """
    
    def __init__(self, script=0, bold=False, italic=False):
        self.script = script
        self.bold = bold
        self.italic = italic
    
    def __add__(self, other):
        # allow None
        if other is None:
            return self
        # set script
        script = other.script
        if script == 0:
            script = self.script
        # done
        return MiniStyle(   script,
                            self.bold or other.bold,
                            self.italic or other.italic )
    
    def __repr__(self):
        tmp = self.script, self.bold, self.italic
        return '<MiniStyle script:%i, bold:%i, italic:%i>' % tmp



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module baseWibjects

Defines the Box class and the DraggableBox class.

"""

import OpenGL.GL as gl

from visvis.core import Wibject
from visvis.core import misc
from visvis.utils.pypoints import Pointset


class Box(Wibject):
    """ Box(parent)
    
    A simple, multi-purpose, rectangle object.
    It implements functionality to draw itself. Most wibjects will
    actually inherit from Box, rather than from Wibject.
    
    """
    
    def __init__(self, parent):
        Wibject.__init__(self, parent)
        self._edgeColor = (0,0,0)
        self._edgeWidth = 1.0
    
    @misc.PropWithDraw
    def edgeColor():
        """ Get/Set the edge color of the wibject.
        """
        def fget(self):
            return self._edgeColor
        def fset(self, value):
            self._edgeColor = misc.getColor(value, 'setting edgeColor')
        return locals()
    
    @misc.PropWithDraw
    def edgeWidth():
        """ Get/Set the edge width of the wibject.
        """
        def fget(self):
            return self._edgeWidth
        def fset(self, value):
            self._edgeWidth = float(value)
        return locals()
    
    def _GetBgcolorToDraw(self):
        """ Can be overloaded to indicate mouse over in buttons.
        """
        return self._bgcolor
    
    def OnDraw(self, fast=False):
        
        # get dimensions
        w, h = self.position.size
        
        # draw plane
        if self._bgcolor:
            # Get positions
            x1, x2 = 0, w
            y1, y2 = 0, h
            # Set color
            clr = self._GetBgcolorToDraw()
            gl.glColor(clr[0], clr[1], clr[2], 1.0)
            #
            gl.glBegin(gl.GL_POLYGON)
            gl.glVertex2f(x1,y1)
            gl.glVertex2f(x1,y2)
            gl.glVertex2f(x2,y2)
            gl.glVertex2f(x2,y1)
            gl.glEnd()
        
        # prepare
        gl.glDisable(gl.GL_LINE_SMOOTH)
        
        # draw edges
        if self.edgeWidth and self.edgeColor:
            
            # Get positions
            # Draw edges on top of the first and last pixel
            x1, x2 = 0.5, w-0.5
            y1, y2 = 0.5, h-0.5
            # Set color and line width
            clr = self.edgeColor
            gl.glColor(clr[0], clr[1], clr[2], 1.0)
            gl.glLineWidth(self.edgeWidth)
            #
            gl.glBegin(gl.GL_LINE_LOOP)
            gl.glVertex2f(x1,y1)
            gl.glVertex2f(x1,y2)
            gl.glVertex2f(x2,y2)
            gl.glVertex2f(x2,y1)
            gl.glEnd()
        
        # clean up
        gl.glEnable(gl.GL_LINE_SMOOTH)


class DraggableBox(Box):
    """ DraggableBox(parent)
    
    A Box wibject, but draggable and resizable.
    Intended as a base class.
    
    """
    
    def __init__(self, parent):
        Box.__init__(self, parent)
        
        # Make me draggable
        self._dragStartPos = None
        self._dragResizing = False
        self._dragMouseOver = False
        
        # Prepare points to draw
        self._DragCalcDots()
        
        # Bind to own events
        self.eventMouseDown.Bind(self._DragOnDown)
        self.eventMouseUp.Bind(self._DragOnUp)
        self.eventEnter.Bind(self._DragOnEnter)
        self.eventLeave.Bind(self._DragOnLeave)
        self.eventPosition.Bind(self._DragCalcDots)
        
        # Bind to figure events
        self.eventMotion.Bind(self._DragOnMove)
    
    
    def _DragCalcDots(self, event=None):
        w,h = self.position.size
        dots = Pointset(2)
        #
        dots.append(3,3); dots.append(3,6); dots.append(3,9)
        dots.append(6,3); dots.append(6,6); dots.append(6,9)
        dots.append(9,3); dots.append(9,6); dots.append(9,9)
        #
        dots.append(w-3, h-3); dots.append(w-3, h-6); dots.append(w-3, h-9)
        dots.append(w-6, h-3); dots.append(w-6, h-6)
        dots.append(w-9, h-3)
        self._dots = dots
    
    def _DragOnEnter(self, event):
        self._dragMouseOver = True
        self.Draw()
    
    def _DragOnLeave(self, event):
        self._dragMouseOver = False
        self.Draw()
   
    def _DragOnDown(self, event):
        pos = self.position
        # Store position if clicked on draggable arreas
        if event.x < 10 and event.y < 10:
            self._dragStartPos = event.absx, event.absy
        elif event.x > pos.width-10 and event.y > pos.height-10:
            self._dragStartPos = event.absx, event.absy
            self._dragResizing = True
    
    
    def _DragOnMove(self, event):
        if not self._dragStartPos:
            return
        elif self._dragResizing:
            self.position.w += event.absx - self._dragStartPos[0]
            self.position.h += event.absy - self._dragStartPos[1]
            event.owner.Draw()
        else: # dragging
            self.position.x += event.absx - self._dragStartPos[0]
            self.position.y += event.absy - self._dragStartPos[1]
            event.owner.Draw()
        self._dragStartPos = event.absx, event.absy
    
    
    def _DragOnUp(self, event):
        self._dragStartPos = None
        self._dragResizing = False
    
    def OnDraw(self):
        Box.OnDraw(self)
        
        if self._dragMouseOver:
            
            # Prepare
            gl.glColor(0,0,0,1)
            gl.glPointSize(1)
            gl.glDisable(gl.GL_POINT_SMOOTH)
            
            # Draw dots
            if len(self._dots):
                gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
                gl.glVertexPointerf(self._dots.data)
                gl.glDrawArrays(gl.GL_POINTS, 0, len(self._dots))
                gl.glDisableClientState(gl.GL_VERTEX_ARRAY)



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

def reportIssue():
    """ help()
    
    Open a webbrowser with the visvis website at the issue list.
    
    
    """
    import webbrowser
    webbrowser.open("https://github.com/almarklein/visvis/issues")

if __name__ == '__main__':
    import visvis as vv
    vv.reportIssue()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv


def ssdfWrite(fname, mesh, name='', bin='unused'):
    """ Simple function that writes a mesh to the ssdf file format.
    """
    
    # Create structure
    s = vv.ssdf.new()
    
    # Populate structure
    s.name = name
    s.vertices = mesh._vertices
    s.normals = mesh._normals
    s.values = mesh._values
    s.faces = mesh._faces
    s.verticesPerFace = mesh._verticesPerFace
    
    # Write
    vv.ssdf.save(fname, s)


def meshWrite(fname, mesh, name='', bin=True):
    """ meshWrite(fname, mesh, name='', bin=True)
    
    Parameters
    ----------
    fname : string
        The filename to write to. The extension should be one of the
        following: .obj .stl .ssdf .bsdf
    mesh : vv.BaseMesh
        The mesh instance to write.
    name : string (optional)
        The name of the object (e.g. 'teapot')
    bin : bool
        For the STL format: whether to write binary, which is much
        more compact then ascii.
    
    Notes on formats
    ----------------
      * The STL format (.stl) is rather limited in the definition of the
        faces; smooth shading is not possible on an STL mesh.
      * The Wavefront format (.obj) is widely available.
      * The SSDF format (.ssdf or .bsdf) is the most efficient in terms
        of memory and speed, but is not widely available.
    
    """
    
    # Use file extension to read file
    if fname.lower().endswith('.stl'):
        import visvis.vvio
        writeFunc = visvis.vvio.stl.StlWriter.write
    elif fname.lower().endswith('.obj'):
        import visvis.vvio
        writeFunc = visvis.vvio.wavefront.WavefrontWriter.write
    elif fname.lower().endswith('.ssdf') or fname.lower().endswith('.bsdf'):
        writeFunc = ssdfWrite
    else:
        raise ValueError('meshWrite cannot determine file type.')
    
    # Read
    return writeFunc(fname, mesh, name, bin)


if __name__ == '__main__':
    bm = vv.meshRead('bunny.ssdf')
    vv.meshWrite('bunny.obj', bm)



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module misc

Various things are defined here that did not fit nicely in any
other module.

This module is also meant to be imported by many
other visvis modules, and therefore should not depend on other
visvis modules.

"""

import os
import sys
import zipfile

import numpy as np
import OpenGL.GL as gl

from visvis.utils import ssdf
from visvis.utils.pypoints import getExceptionInstance  # noqa


V2 = sys.version_info[0] == 2
if V2:
    unichr = unichr  # noqa
    basestring = basestring  # noqa
else:
    basestring = str
    unichr = chr


# Used to load resources when in a zipfile
def splitPathInZip(path):
    """ splitPathInZip(path)
    Split a filename in two parts: the path to a zipfile, and the path
    within the zipfile. If the given path is a native file or direcory,
    returns ('', path). Raises an error if no valid zipfile can be found.
    """
    ori_path = path
    if os.path.exists(path):
        return '', path
    # Split path in parts
    zipPath = []
    while True:
        path, sub = os.path.split(path)
        if not sub:
            raise RuntimeError('Not a real path nor in a zipfile: "%s"' % ori_path)
        zipPath.insert(0, sub)
        if os.path.isfile(path):
            if zipfile.is_zipfile(path):
                return path, os.path.join(*zipPath)
            else:
                raise RuntimeError('Not a zipfile: "%s"' % ori_path)
    


## For info about OpenGL version


# To store openGl info
_glInfo = [None]*4


def ensureString(s):
    if isinstance(s, str):
        return s
    else:
        return s.decode('ascii')

def getOpenGlInfo():
    """ getOpenGlInfo()
    
    Get information about the OpenGl version on this system.
    Returned is a tuple (version, vendor, renderer, extensions)
    Note that this function will return 4 Nones if the openGl
    context is not set.
    
    """
    
    if gl.glGetString(gl.GL_VERSION) is None:
        raise RuntimeError('There is currently no OpenGL context')
    
    if not _glInfo[0]:
        _glInfo[0] = ensureString(gl.glGetString(gl.GL_VERSION))
        _glInfo[1] = ensureString(gl.glGetString(gl.GL_VENDOR))
        _glInfo[2] = ensureString(gl.glGetString(gl.GL_RENDERER))
        _glInfo[3] = ensureString(gl.glGetString(gl.GL_EXTENSIONS))
    return tuple(_glInfo)

_glLimitations = {}
def getOpenGlCapable(version, what=None):
    """ getOpenGlCapable(version, what)
    
    Returns True if the OpenGl version on this system is equal or higher
    than the one specified and False otherwise.
    
    If False, will display a message to inform the user, but only the first
    time that this limitation occurs (identified by the second argument).
    
    """
    
    # obtain version of system
    curVersion = _glInfo[0]
    if not curVersion:
        curVersion, dum1, dum2, dum3 = getOpenGlInfo()
        if not curVersion:
            return False # OpenGl context not set, better safe than sory
    
    # make sure version is a string
    if isinstance(version, (int,float)):
        version = str(version)
    
    # test
    if curVersion >= version :
        return True
    else:
        # show message?
        if what and (what not in _glLimitations):
            _glLimitations[what] = True
            tmp = "Warning: the OpenGl version on this system is too low "
            tmp += "to support " + what + ". "
            tmp += "Try updating your drivers or buy a new video card."
            print(tmp)
        return False


## Decorators


def Property(function):
    """ Property(function)
    
    A property decorator which allows to define fget, fset and fdel
    inside the function.
    
    Note that the class to which this is applied must inherit from object!
    Code based on an example posted by Walker Hale:
    http://code.activestate.com/recipes/410698/#c6
    
    Example
    -------
    
    class Example(object):
        @Property
        def myattr():
            ''' This is the doc string.
            '''
            def fget(self):
                return self._value
            def fset(self, value):
                self._value = value
            def fdel(self):
                del self._value
            return locals()
    
    """
    
    # Define known keys
    known_keys = 'fget', 'fset', 'fdel', 'doc'
    
    # Get elements for defining the property. This should return a dict
    func_locals = function()
    if not isinstance(func_locals, dict):
        raise RuntimeError('Property function should "return locals()".')
    
    # Create dict with kwargs for property(). Init doc with docstring.
    D = {'doc': function.__doc__}
    
    # Copy known keys. Check if there are invalid keys
    for key in list(func_locals.keys()):
        if key in known_keys:
            D[key] = func_locals[key]
        else:
            raise RuntimeError('Invalid Property element: %s' % key)
    
    # Done
    return property(**D)


def PropWithDraw(function):
    """ PropWithDraw(function)
    
    A property decorator which allows to define fget, fset and fdel
    inside the function.
    
    Same as Property, but calls self.Draw() when using fset.
    
    """
    
    # Define known keys
    known_keys = 'fget', 'fset', 'fdel', 'doc'
    
    # Get elements for defining the property. This should return a dict
    func_locals = function()
    if not isinstance(func_locals, dict):
        raise RuntimeError('Property function should "return locals()".')
    
    # Create dict with kwargs for property(). Init doc with docstring.
    D = {'doc': function.__doc__}
    
    # Copy known keys. Check if there are invalid keys
    for key in list(func_locals.keys()):
        if key in known_keys:
            D[key] = func_locals[key]
        else:
            raise RuntimeError('Invalid Property element: %s' % key)
    
    # Replace fset
    fset = D.get('fset', None)
    def fsetWithDraw(self, *args):
        fset(self, *args)
        if hasattr(self, 'Draw'):
            self.Draw()
        #print(fset._propname, self, time.time())
    if fset:
        fset._propname = function.__name__
        D['fset'] = fsetWithDraw
    
    # Done
    return property(**D)
    

def DrawAfter(function):
    """ DrawAfter(function)
    
    Decorator for methods that make self.Draw() be called right after
    the function is called.
    
    """
    def newFunc(self, *args, **kwargs):
        retval = function(self, *args, **kwargs)
        if hasattr(self, 'Draw'):
            self.Draw()
        return retval
    newFunc.__doc__ = function.__doc__
    return newFunc


def PropertyForSettings(function):
    """ PropertyForSettings(function)
    
    A property decorator that also supplies the function name to the
    fget and fset function. The fset method also calls _Save()
    
    """
    
    # Define known keys
    known_keys = 'fget', 'fset', 'fdel', 'doc'
    
    # Get elements for defining the property. This should return a dict
    func_locals = function()
    if not isinstance(func_locals, dict):
        raise RuntimeError('From visvis version 1.6, Property function should "return locals()".')
    
    # Create dict with kwargs for property(). Init doc with docstring.
    D = {'doc': function.__doc__}
    
    # Copy known keys. Check if there are invalid keys
    for key in list(func_locals.keys()):
        if key in known_keys:
            D[key] = func_locals[key]
        else:
            raise RuntimeError('Invalid Property element: %s' % key)
    
    # Replace fset and fget
    fset = D.get('fset', None)
    fget = D.get('fget', None)
    def fsetWithKey(self, *args):
        fset(self, function.__name__, *args)
        self._Save()
    def fgetWithKey(self, *args):
        return fget(self, function.__name__)
    if fset:
        D['fset'] = fsetWithKey
    if fget:
        D['fget'] = fgetWithKey
    
    # Done
    return property(**D)


## The range class


class Range(object):
    """ Range(min=0, max=0)
    
    Represents a range (a minimum and a maximum ). Can also be instantiated
    using a tuple.
    
    If max is set smaller than min, the min and max are flipped.
    
    """
    def __init__(self, min=0, max=1):
        self.Set(min,max)
    
    def Set(self, min=0, max=1):
        """ Set the values of min and max with one call.
        Same signature as constructor.
        """
        if isinstance(min, Range):
            min, max = min.min, min.max
        elif isinstance(min, (tuple,list)):
            min, max = min[0], min[1]
        self._min = float(min)
        self._max = float(max)
        self._Check()
    
    @property
    def range(self):
        return self._max - self._min
    
    @Property # visvis.Property
    def min():
        """ Get/Set the minimum value of the range. """
        def fget(self):
            return self._min
        def fset(self,value):
            self._min = float(value)
            self._Check()
        return locals()
    
    @Property # visvis.Property
    def max():
        """ Get/Set the maximum value of the range. """
        def fget(self):
            return self._max
        def fset(self,value):
            self._max = float(value)
            self._Check()
        return locals()
    
    def _Check(self):
        """ Flip min and max if order is wrong. """
        if self._min > self._max:
            self._max, self._min = self._min, self._max
    
    def Copy(self):
        return Range(self.min, self.max)
        
    def __repr__(self):
        return "<Range %1.2f to %1.2f>" % (self.min, self.max)


## Transform classes for wobjects
    
    
class Transform_Base(object):
    """ Transform_Base
    
    Base transform object.
    Inherited by classes for translation, scale and rotation.
    
    """
    pass
    
class Transform_Translate(Transform_Base):
    """ Transform_Translate(dx=0.0, dy=0.0, dz=0.0)
    
    Translates the wobject.
    
    """
    def __init__(self, dx=0.0, dy=0.0, dz=0.0):
        self.dx = dx
        self.dy = dy
        self.dz = dz

class Transform_Scale(Transform_Base):
    """ Transform_Scale(sx=1.0, sy=1.0, sz=1.0)
    
    Scales the wobject.
    
    """
    def __init__(self, sx=1.0, sy=1.0, sz=1.0):
        self.sx = sx
        self.sy = sy
        self.sz = sz

class Transform_Rotate(Transform_Base):
    """ Transform_Rotate( angle=0.0, ax=0, ay=0, az=1, angleInRadians=None)
    
    Rotates the wobject. Angle is in degrees.
    Use angleInRadians to specify the angle in radians,
    which is then converted in degrees.
    """
    def __init__(self, angle=0.0, ax=0, ay=0, az=1, angleInRadians=None):
        if angleInRadians is not None:
            angle = angleInRadians * 180 / np.pi
        self.angle = angle
        self.ax = ax
        self.ay = ay
        self.az = az

## Colour stuff

# Define named colors
colorDict = {}
colorDict['black']  = colorDict['k'] = (0,0,0)
colorDict['white']  = colorDict['w'] = (1,1,1)
colorDict['red']    = colorDict['r'] = (1,0,0)
colorDict['green']  = colorDict['g'] = (0,1,0)
colorDict['blue']   = colorDict['b'] = (0,0,1)
colorDict['cyan']   = colorDict['c'] = (0,1,1)
colorDict['yellow'] = colorDict['y'] = (1,1,0)
colorDict['magenta']= colorDict['m'] = (1,0,1)

def getColor(value, descr='getColor'):
    """ getColor(value, descr='getColor')
    
    Make sure a value is a color. If a character is given, returns the color
    as a tuple.
    
    """
    tmp = ""
    if not value:
        value = None
    elif isinstance(value, basestring):
        if value not in colorDict:
            tmp = "string color must be one of 'rgbycmkw' !"
        else:
            value = colorDict[value]
    elif isinstance(value, (list, tuple)):
        if len(value) != 3:
            tmp = "tuple color must be length 3!"
        value = tuple(value)
    else:
        tmp = "color must be a three element tuple or a character!"
    # error or ok?
    if tmp:
        raise ValueError("Error in %s: %s" % (descr, tmp) )
    return value


## More functions ...

def isFrozen():
    """ isFrozen
    
    Returns whether this is a frozen application
    (using bbfreeze or py2exe). From pyzolib.paths.py
    
    """
    return bool( getattr(sys, 'frozen', None) )


# todo: cx_Freeze and friends should provide a mechanism to store
# resources automatically ...
def getResourceDir():
    """ getResourceDir()
    
    Get the directory to the visvis resources.
    
    """
    if isFrozen():
        # See application_dir() in pyzolib/paths.py
        path =  os.path.abspath(os.path.dirname(sys.path[0]))
    else:
        path = os.path.abspath( os.path.dirname(__file__) )
        path = os.path.split(path)[0]
    return os.path.join(path, 'visvisResources')


# From pyzolib/paths.py
import os, sys  # noqa
def appdata_dir(appname=None, roaming=False, macAsLinux=False):
    """ appdata_dir(appname=None, roaming=False,  macAsLinux=False)
    Get the path to the application directory, where applications are allowed
    to write user specific files (e.g. configurations). For non-user specific
    data, consider using common_appdata_dir().
    If appname is given, a subdir is appended (and created if necessary).
    If roaming is True, will prefer a roaming directory (Windows Vista/7).
    If macAsLinux is True, will return the Linux-like location on Mac.
    """
    
    # Define default user directory
    userDir = os.path.expanduser('~')
    
    # Get system app data dir
    path = None
    if sys.platform.startswith('win'):
        path1, path2 = os.getenv('LOCALAPPDATA'), os.getenv('APPDATA')
        path = (path2 or path1) if roaming else (path1 or path2)
    elif sys.platform.startswith('darwin') and not macAsLinux:
        path = os.path.join(userDir, 'Library', 'Application Support')
    # On Linux and as fallback
    if not (path and os.path.isdir(path)):
        path = userDir
    
    # Maybe we should store things local to the executable (in case of a
    # portable distro or a frozen application that wants to be portable)
    prefix = sys.prefix
    if getattr(sys, 'frozen', None): # See application_dir() function
        prefix = os.path.abspath(os.path.dirname(sys.path[0]))
    for reldir in ('settings', '../settings'):
        localpath = os.path.abspath(os.path.join(prefix, reldir))
        if os.path.isdir(localpath):
            try:
                open(os.path.join(localpath, 'test.write'), 'wb').close()
                os.remove(os.path.join(localpath, 'test.write'))
            except IOError:
                pass # We cannot write in this directory
            else:
                path = localpath
                break
    
    # Get path specific for this app
    if appname:
        if path == userDir:
            appname = '.' + appname.lstrip('.') # Make it a hidden directory
        path = os.path.join(path, appname)
        if not os.path.isdir(path):
            os.mkdir(path)
    
    # Done
    return path


class Settings(object):
    """ Global settings object.
    
    This object can be used to set the visvis settings in an easy way
    from the Python interpreter.
    
    The settings are stored in a file in the user directory (the filename
    can be obtained using the _fname attribute).
    
    Note that some settings require visvis to restart.
    
    """
    def __init__(self):
        
        # Define settings file name
        self._fname = os.path.join(appdata_dir('visvis'), 'config.ssdf')
        
        # Init settings
        self._s = ssdf.new()
        
        # Load settings if we can
        if os.path.exists(self._fname):
            try:
                self._s = ssdf.load(self._fname)
            except Exception:
                pass
        
        # Update any missing settings to their defaults
        for key in dir(self):
            if key.startswith('_'):
                continue
            self._s[key] = getattr(self, key)
        
        # Save now so the config file contains all settings
        self._Save()
    
    def _Save(self):
        try:
            ssdf.save(self._fname, self._s)
        except IOError:
            pass # Maybe an installed frozen application (no write-rights)
    
    @PropertyForSettings
    def preferredBackend():
        """ The preferred backend GUI toolkit to use
        ('pyside', 'pyqt4', 'wx', 'gtk', 'fltk').
          * If the selected backend is not available, another one is selected.
          * If preferAlreadyLoadedBackend is True, will prefer a backend that
            is already imported.
          * Requires a restart.
        """
        def fget(self, key):
            if key in self._s:
                return self._s[key]
            else:
                return 'pyside'  # Default value
        def fset(self, key, value):
            # Note that 'qt4' in valid for backward compatibility
            value = value.lower()
            if value in ['pyside', 'qt4', 'pyqt4', 'wx', 'gtk', 'fltk', 'foo']:
                self._s[key] = value
            else:
                raise ValueError('Invalid backend specified.')
        return locals()
    
    @PropertyForSettings
    def preferAlreadyLoadedBackend():
        """ Bool that indicates whether visvis should prefer an already
        imported backend (even if it's not the preferredBackend). This is
        usefull in interactive session in for example IEP, Spyder or IPython.
        Requires a restart.
        """
        def fget(self, key):
            if key in self._s:
                return bool(self._s[key])
            else:
                return True  # Default value
        def fset(self, key, value):
            self._s[key] = bool(value)
        return locals()
    
#     @PropertyForSettings
#     def defaultInterpolation2D():
#         """ The default interpolation mode for 2D textures (bool). If False
#         the pixels are well visible, if True the image looks smoother.
#         Default is False.
#         """
#         def fget(self, key):
#             if key in self._s:
#                 return bool(self._s[key])
#             else:
#                 return False  # Default value
#         def fset(self, key, value):
#             self._s[key] = bool(value)
#         return locals()
    
    @PropertyForSettings
    def figureSize():
        """ The initial size for figure windows. Should be a 2-element
        tuple or list. Default is (560, 420).
        """
        def fget(self, key):
            if key in self._s:
                return tuple(self._s[key])
            else:
                return (560, 420)  # Default value
        def fset(self, key, value):
            if not isinstance(value, (list,tuple)) or len(value) != 2:
                raise ValueError('Figure size must be a 2-element list or tuple.')
            value = [int(i) for i in value]
            self._s[key] = tuple(value)
        return locals()
    
    @PropertyForSettings
    def volshowPreference():
        """ Whether the volshow() function prefers the volshow2() or volshow3()
        function. By default visvis prefers volshow3(), but falls back to
        volshow2() when the OpenGl version is not high enough. Some OpenGl
        drivers, however, support volume rendering only in ultra-slow software
        mode (seen on ATI). In this case, or if you simply prefer volshow2()
        you can set this setting to '2'.
        """
        def fget(self, key):
            if key in self._s:
                return self._s[key]
            else:
                return 3  # Default value
        def fset(self, key, value):
            if value not in [2,3]:
                raise ValueError('volshowPreference must be 2 or 3.')
            self._s[key] = int(value)
        return locals()
    
    @PropertyForSettings
    def defaultFontName():
        """ The default font to use. Can be 'mono', 'sans' or 'serif', with
        'sans' being the default.
        """
        def fget(self, key):
            if key in self._s:
                return self._s[key]
            else:
                return 'sans'  # Default value
        def fset(self, key, value):
            value = value.lower()
            if value not in ['mono', 'sans', 'serif', 'humor']:
                raise ValueError("defaultFontName must be 'mono', 'sans', 'serif' or 'humor'.")
            self._s[key] = value
        return locals()
    
    @PropertyForSettings
    def defaultRelativeFontSize():
        """ The default relativeFontSize of new figures. The relativeFontSize
        property can be used to scale all text simultenously, as well as
        increase/decrease the margins availavle for the text. The default is
        1.0.
        """
        def fget(self, key):
            if key in self._s:
                return self._s[key]
            else:
                return 1.0
        def fset(self, key, value):
            self._s[key] = float(value)
        return locals()
    
    # todo: more? maybe axes bgcolor and axisColor?

# Create settings instance, this is what gets inserted in the visvis namespace
settings = Settings()

# Set __file__ absolute when loading
__file__ = os.path.abspath(__file__)



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import time
import visvis as vv

def callLater(delay, callable, *args, **kwargs):
    """ callLater(delay, callable, *args, **kwargs)
    
    Call a callable after a specified delay (in seconds),
    with the specified arguments and keyword-arguments.
    
    Parameters
    ----------
    delay : scalar
        The delay in seconds. If zero, the callable is called right
        after the current processing has returned to the main loop,
        before any other visvis events are processed.
    callable : a callable object
        The callback that is called when the delay has passed.
    
    See also vv.Event
    
    """
    calltime = time.time() + delay
    vv.events._callLater_callables[calltime]= (callable, args, kwargs)

if __name__ == '__main__':
    # Code to call a method after 2.0 seconds
    def foo():
        print('haha!')
    vv.callLater(2.0, foo)
    # For this to work, a visvis application must be Create'd.
    app = vv.use()
    app.Create()
    app.Run()



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2017, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv


def isocontour(im, isovalue=None):
    """ isocontour(im, isovalue=None)
    
    Uses scikit-image to calculate the iso contours for the given 2D
    image. If isovalue is not given or None, a value between the min
    and max of the image is used.
    
    Returns a pointset in which each two subsequent points form a line
    piece. This van be best visualized using "vv.plot(result, ls='+')".
    
    """
    
    from visvis.utils.iso import isocontour as _isocontour
    
    return _isocontour(im, isovalue)


if __name__ == '__main__':
    vv.plot(isocontour(vv.imread('astronaut.png')[:,:,1]), ls='+', lw=3)



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv

def clf():
    """ clf()
    
    Clear current figure.
    
    """
    f = vv.gcf()
    f.Clear()
    return f


if __name__ == '__main__':
    # Show plot
    vv.plot([1,2,3,1,4])
    # Clear the whole figure
    vv.clf()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Keith Smith
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.
#
# Thanks to Keith Smith for implementing the polar plot functionality.

import numpy as np
import visvis as vv

from visvis.utils.pypoints import Pointset, is_Point, is_Pointset
from visvis import PolarLine


def makeArray(data):
    if isinstance(data, np.ndarray):
        return data
    else:
        # create numpy array
        try:
            l = len(data)
            a = np.empty((l, 1))
            for i in range(len(data)):
                a[i] = data[i]
            return a
        except TypeError:
            raise Exception("Cannot plot %s" % data.__class__.__name__)


def _SetLimitsAfterDraw(event):
    """ To be able to set the limits after the first draw. """
    # Set limits
    fig = event.owner
    for axis in fig.FindObjects(vv.axises.PolarAxis2D):
        limits = axis.GetLimits()
        axis.SetLimits(rangeTheta=limits[0], rangeR=limits[1])
    # Unsubscribe and redraw
    fig.eventAfterDraw.Unbind(_SetLimitsAfterDraw)
    fig.Draw()

    
def polarplot(data1, data2=None, inRadians=False,
            lw=1, lc='b', ls="-", mw=7, mc='b', ms='', mew=1, mec='k',
            alpha=1, axesAdjust=True, axes=None, **kwargs):
    """ polarplot(*args, inRadians=False,
            lw=1, lc='b', ls="-", mw=7, mc='b', ms='', mew=1, mec='k',
            alpha=1, axesAdjust=True, axes=None):
    
    Plot 2D polar data, using a polar axis to draw a polar grid.
    
    Usage
    -----
      * plot(Y, ...) plots a 1D polar signal.
      * plot(X, Y, ...) also supplies angular coordinates
      * plot(P, ...) plots using a Point or Pointset instance
    
    Keyword arguments
    -----------------
    (The longer names for the line properties can also be used)
    lw : scalar
        lineWidth. The width of the line. If zero, no line is drawn.
    mw : scalar
        markerWidth. The width of the marker. If zero, no marker is drawn.
    mew : scalar
        markerEdgeWidth. The width of the edge of the marker.
    lc : 3-element tuple or char
        lineColor. The color of the line. A tuple should represent the RGB
        values between 0 and 1. If a char is given it must be
        one of 'rgbmcywk', for reg, green, blue, magenta, cyan, yellow,
        white, black, respectively.
    mc : 3-element tuple or char
        markerColor. The color of the marker. See lineColor.
    mec : 3-element tuple or char
        markerEdgeColor. The color of the edge of the marker.
    ls : string
        lineStyle. The style of the line. (See below)
    ms : string
        markerStyle. The style of the marker. (See below)
    axesAdjust : bool
        If axesAdjust==True, this function will call axes.SetLimits(), and set
        the camera type to 2D.
    axes : Axes instance
        Display the image in this axes, or the current axes if not given.
    
    Line styles
    -----------
      * Solid line: '-'
      * Dotted line: ':'
      * Dashed line: '--'
      * Dash-dot line: '-.' or '.-'
      * A line that is drawn between each pair of points: '+'
      * No line: '' or None.
    
    Marker styles
    -------------
      * Plus: '+'
      * Cross: 'x'
      * Square: 's'
      * Diamond: 'd'
      * Triangle (pointing up, down, left, right): '^', 'v', '<', '>'
      * Pentagram star: 'p' or '*'
      * Hexgram: 'h'
      * Point/cirle: 'o' or '.'
      * No marker: '' or None
    
    Polar axis
    ----------
    This polar axis has a few specialized methods for adjusting the polar
    plot. Access these via vv.gca().axis.
      * SetLimits(thetaRange, radialRange)
      * thetaRange, radialRange = GetLimits()
      * angularRefPos: Get and Set methods for the relative screen
        angle of the 0 degree polar reference.  Default is 0 degs
        which corresponds to the positive x-axis (y =0)
      * isCW: Get and Set methods for the sense of rotation CCW or
        CW. This method takes/returns a bool (True if the default CW).
    
    Interaction
    -----------
      * Drag mouse up/down to translate radial axis.
      * Drag mouse left/right to rotate angular ref position.
      * Drag mouse + shift key up/down to rescale radial axis (min R fixed).
    
    """

    # create a dict from the properties and combine with kwargs
    tmp = {'lineWidth': lw, 'lineColor': lc, 'lineStyle': ls,
                'markerWidth': mw, 'markerColor': mc, 'markerStyle': ms,
                'markerEdgeWidth': mew, 'markerEdgeColor': mec}
    for i in tmp:
        if not i in kwargs:
            kwargs[i] = tmp[i]
    
    ##  create the data
    if is_Pointset(data1):
        pp = data1
    elif is_Point(data1):
        pp = Pointset(data1.ndim)
        pp.append(data1)
    else:

        if data1 is None:
            raise ValueError("The first argument cannot be None!")
        data1 = makeArray(data1)
        
        if data2 is None:
            # R data is given, thetadata must be
            # a range starting from 0 degrees
            data2 = data1
            data1 = np.arange(0, data2.shape[0])
        else:
            data2 = makeArray(data2)
        
        # check dimensions
        L = data1.size
        if L != data2.size:
            raise ValueError("Array dimensions do not match! %i vs %i " %
                    (data1.size, data2.size))
        
        # build points
        data1 = data1.reshape((data1.size, 1))
        data2 = data2.reshape((data2.size, 1))

    if not inRadians:
        data1 = np.pi * data1 / 180.0

    ## create the line
    if axes is None:
        axes = vv.gca()
    axes.axisType = 'polar'
    fig = axes.GetFigure()
    
    l = PolarLine(axes, data1, data2)
    l.lw = kwargs['lineWidth']
    l.lc = kwargs['lineColor']
    l.ls = kwargs['lineStyle']
    l.mw = kwargs['markerWidth']
    l.mc = kwargs['markerColor']
    l.ms = kwargs['markerStyle']
    l.mew = kwargs['markerEdgeWidth']
    l.mec = kwargs['markerEdgeColor']
    l.alpha = alpha

    ## almost done...
    
    # Init axis
#     axes.axis.SetLimits()
    
    if axesAdjust:
        if axes.daspectAuto is None:
            axes.daspectAuto = True
        axes.cameraType = '2d'
        axes.SetLimits()
    
    # Subsribe after-draw event handler
    # (unsubscribe first in case we do multiple plots)
    fig.eventAfterDraw.Unbind(_SetLimitsAfterDraw)
    fig.eventAfterDraw.Bind(_SetLimitsAfterDraw)
    
    # Return
    axes.Draw()
    return l


if __name__ == '__main__':
    # Make data
    angs = 0.1 + np.linspace(-90, 90, 181)  # 0.1+ get rid of singularity
    angsRads = np.pi * angs / 180.0
    mag = 10 * np.log10(np.abs(np.sin(10 * angsRads) / angsRads)) + angsRads
    mag = mag - np.max(mag)
    # Show data
    vv.polarplot( angs, mag, lc='b')
    vv.polarplot(angs+20, mag, lc='r', lw=2)
    a = vv.gca() # Triggers an update required for polar plots



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module line

Defines the Line class, which represents a line connected (optionally)
with markers. It is the object created with the plot() function.

The lines are drawn simply with OpenGL lines. The markers are drawn
with OpenGl points if possible, and using sprites otherwise.

Since this is such a fundamental part of visvis, and it's uses by
for example the Legend class, this module is part of the core.

"""

import OpenGL.GL as gl
import numpy as np
import math

from visvis.utils.pypoints import Pointset, is_Pointset
from visvis.core.misc import PropWithDraw, DrawAfter, basestring
from visvis.core.misc import Range, getColor, getOpenGlCapable
from visvis.core.base import Wobject


# int('1010101010101010',2)  int('1100110011001100',2)
lineStyles = {  ':':int('1010101010101010',2),  '--':int('1111000011110000',2),
                '-.':int('1110010011100100',2), '.-':int('1110010011100100',2),
                '-':False, '+':False}


class Sprite:
    """ Sprite(data, width)
    
    Represents an OpenGL sprite object.
    
    """

    def __init__(self, data, width):
        """ Supply the data, which must be uint8 alpha data,
        preferably shaped with a power of two. """
        self._texId = 0
        self._data = data
        self._width = width
        self._canUse = None # set to True/False if OpenGl version high enough

    def Create(self):
        """ Create an OpenGL texture from the data. """
        
        # detemine now if we can use point sprites
        self._canUse = getOpenGlCapable('2.0',
            'point sprites (for advanced markers)')
        if not self._canUse:
            return
        
        # gl.glEnable(gl.GL_TEXTURE_2D)
        
        # make texture
        self._texId = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._texId)
        
        # set interpolation and extrapolation parameters
        tmp = gl.GL_NEAREST # gl.GL_NEAREST | gl.GL_LINEAR
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, tmp)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, tmp)

        # upload data
        shape = self._data.shape
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_ALPHA, shape[0], shape[1],
            0, gl.GL_ALPHA, gl.GL_UNSIGNED_BYTE, self._data)

    
    @property
    def usable(self):
        return self._canUse
    
    
    def Enable(self):
        """ Enable the sprite, drawing points after calling this
        draws this sprite at each point. """
        
        if not self._texId and self._canUse in [None, True]:
            self.Create()
        
        if not self._canUse:
            gl.glEnable(gl.GL_POINT_SMOOTH)
            gl.glPointSize(self._width)
            return # proceed if None; canUse is assigned in Create()
        
        # bind to texture
        gl.glEnable(gl.GL_TEXTURE_2D)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._texId)

        # get allowed point size
        sizeRange = gl.glGetFloatv(gl.GL_ALIASED_POINT_SIZE_RANGE)
        gl.glPointParameterf( gl.GL_POINT_SIZE_MIN, sizeRange[0] )
        gl.glPointParameterf( gl.GL_POINT_SIZE_MAX, sizeRange[1] )
        
        # enable sprites and set params
        gl.glEnable(gl.GL_POINT_SPRITE)

        # tell opengl to iterate over the texture
        gl.glTexEnvf( gl.GL_POINT_SPRITE, gl.GL_COORD_REPLACE, gl.GL_TRUE )


    def Disable(self):
        """ Return to normal points. """
        if self._canUse:
            gl.glDisable(gl.GL_TEXTURE_2D)
            gl.glDisable(gl.GL_POINT_SPRITE)
        else:
            gl.glDisable(gl.GL_POINT_SMOOTH)


    def Destroy(self):
        """ Destroy the sprite, removing the texture from opengl. """
        if self._texId > 0:
            try:
                gl.glDeleteTextures([self._texId])
                self._texId = 0
            except Exception:
                pass

    def __del__(self):
        """ Delete when GC cleans up. """
        self.Destroy()


class MarkerManager:
    """ MarkerManager()
    
    The markermanager manages sprites to draw the markers. It
    creates the sprite textures on the fly when they are needed.
    Already created sprites are reused.

    Given the markerStyle, markerWidth and markerEdgeWidth a marker
    can be requested.
    
    """

    def __init__(self):
        # a dict of 3 element tuples (size, sprite1, sprite2)
        # where the first is for the face, the second for the edge.
        self.sprites = {}

    def GetSprites(self, ms, mw, mew):
        """ GetSprites(mw, mw, mew)
        
        Get the sprites for drawing the edge and the face, given
        the ms, mw and mew.
        
        This will create the appropriate sprites or reuse a previously
        created set of sprites if available.
        
        Returns a tuple (size, faceSprite, edgeSprite)
        """
        # find the diameter which best fits, but which is a multiple of 4
        # such that 32 bits are always filled.
        mw, mew = int(mw), int(mew)

        # create key of these settings
        key = "%s_%i_%i" % (ms, mw, mew)

        # if it does not exist, create it!
        if key not in self.sprites:
            self.sprites[key] = self._CreateSprites(ms, mw, mew)
        else:
            # check if it is a valid texture ...
            id = self.sprites[key][1]._texId
            if not gl.glIsTexture(id):
                self.sprites[key] = self._CreateSprites(ms, mw, mew)

        # return
        return self.sprites[key]


    def _CreateSprites(self, ms, mw, mew):
        """ Create the sprites from scratch. """

        ## init
        # We'll make a 2D array of size d, which fits the marker completely.
        # Then we make a second array with the original eroded as many
        # times as the edge is wide.

        # find nearest multiple of four
        d = 4
        while d < mw+2*mew:
            d += 4
        # what is the offset for the face
        #dd = ( d-(mw+2*mew) ) / 2
        dd = (d-mw)//2

        # calc center
        c = mw/2.0

        # create patch
        data1 = np.zeros((d,d),dtype=np.uint8)

        # create subarray for face
        data2 = data1[dd:,dd:][:mw,:mw]

        ## define marker functions
        def square(xy):
            x, y = xy
            data2[y,x]=255
        def diamond(xy):
            x, y = xy
            if y > x-mw/2 and y<x+mw/2 and y > (mw-x)-c and y<(mw-x)+c:
                data2[y,x]=255
        def plus(xy):
            x, y = xy
            if y > mw/3 and y < 2*mw/3:
                data2[y,x]=255
            if x > mw/3 and x < 2*mw/3:
                data2[y,x]=255
        def cross(xy):
            x, y = xy
            if y > x-mw/4 and y < x+mw/4:
                data2[y,x]=255
            if y > (mw-x)-mw/4 and y < (mw-x)+mw/4:
                data2[y,x]=255
        def flower(xy):
            x, y = xy
            a = math.atan2(y-c,x-c)
            r = (x-c)**2 + (y-c)**2
            relAng = 5 * a / (2*math.pi)  # whole circle from 1 to 5
            subAng = (relAng % 1)       # get the non-integer bit
            if subAng>0.5: subAng = 1-subAng
            refRad1, refRad2 = c/4, c
            a = math.sin(subAng*math.pi)
            refRad = (1-a)*refRad1 + a*refRad2
            if r < refRad**2:
                data2[y,x]=255
        def star5(xy):
            x, y = xy
            a = math.atan2(y-c,x-c) - 0.5*math.pi
            r = (x-c)**2 + (y-c)**2
            relAng = 5 * a / (2*math.pi)  # whole circle from 1 to 5
            subAng = (relAng % 1)       # get the non-integer bit
            if subAng>0.5: subAng = 1-subAng
            refRad1, refRad2 = c/4, c
            a = math.asin(subAng*2) / (math.pi/2)
            refRad = (1-a)*refRad1 + a*refRad2
            if r < refRad**2:
                data2[y,x]=255
        def star6(xy):
            x, y = xy
            a = math.atan2(y-c,x-c)
            r = (x-c)**2 + (y-c)**2
            relAng = 6 * a / (2*math.pi)  # whole circle from 1 to 5
            subAng = (relAng % 1)       # get the non-integer bit
            if subAng>0.5: subAng = 1-subAng
            refRad1, refRad2 = c/3, c
            a = math.asin(subAng*2) / (math.pi/2)
            refRad = (1-a)*refRad1 + a*refRad2
            if r < refRad**2:
                data2[y,x]=255
        def circle(xy):
            x,y = xy
            r = (x-c)**2 + (y-c)**2
            if r < c**2:
                data2[y,x] = 255
        def triangleDown(xy):
            x,y = xy
            if x >= 0.5*y and x <= mw-0.5*(y+1):
                data2[y,x] = 255
        def triangleUp(xy):
            x,y = xy
            if x >= c-0.5*y and x <= c+0.5*y:
                data2[y,x] = 255
        def triangleLeft(xy):
            x,y = xy
            if y >= c-0.5*x and y <= c+0.5*x:
                data2[y,x] = 255
        def triangleRight(xy):
            x,y = xy
            print('oef', x, y)
            if y >= 0.5*x and y <= mw-0.5*(x+1):
                data2[y,x] = 255

        # a dict ms to function
        funcs = {   's':square, 'd':diamond, '+':plus, 'x':cross,
                    '*':star5, 'p':star5, 'h':star6, 'f':flower,
                    '.':circle, 'o':circle, 'v':triangleDown,
                    '^':triangleUp, '<':triangleLeft, '>':triangleRight}

        # select function
        try:
            func = funcs[ms]
        except KeyError:
            func = circle
        
        ## Create face
        I,J = np.where(data2==0)
        for xy in zip(I,J):
            func(xy)
        
        ## dilate x times to create edge
        # we add a border to the array to make the dilation possible
        data3 = np.zeros((d+4,d+4), dtype=np.uint8)
        data3[2:-2,2:-2] = 1
        # retrieve indices.
        I,J = np.where(data3==1)
        # copy face
        data3[2:-2,2:-2] = data1
        tmp = data3.copy()
        # apply
        def dilatePixel(xy):
            x,y = xy
            if tmp[y-1:y+2,x-1:x+2].max():
                data3[y,x] = 255
        for i in range(int(mew)):
            for xy in zip(I,J):
                dilatePixel(xy)
            tmp = data3.copy()
        # remove border
        data3 = data3[2:-2,2:-2]

        ## create sprites and return

        sprite1 = Sprite(data1, mw)
        sprite2 = Sprite(data3-data1, mw+2*mew)
        
        return d, sprite1, sprite2


class Line(Wobject):
    """ Line(parent, points)

    The line class represents a set of points (locations) in world coordinates.
    This class can draw lines between the points, markers at the point
    coordinates, or both.
    
    Line objects can be created with the function vv.plot().
    
    Performance tips
    ----------------
    The s, o (and .) styles can be drawn using standard
    OpenGL points if alpha is 1 or if no markeredge is drawn.
    
    Otherwise point sprites are used, which can be slower
    on some (older) cards (like ATI, Nvidia performs quite ok with with
    sprites).
    
    Some video cards simply do not support sprites (seen on ATI).
    
    """

    def __init__(self, parent, points):
        Wobject.__init__(self, parent)
        
        # Store points
        self.SetPoints(points)
        
        # init line properties
        self._lw, self._ls, self._lc = 1, '-', 'b'
        # init marker properties
        self._mw, self._ms, self._mc = 7, '', 'b'
        # init marker edge properties
        self._mew, self._mec = 1, 'k'

        # alpha values
        self._alpha1 = 1


    def _AsFloat(self, value, descr):
        """ Make sure a value is a float. """
        try:
            value = float(value)
            if value<0:
                raise ValueError()
        except ValueError:
            tmp = "the value must be a number equal or larger than zero!"
            raise Exception("Error in %s: %s" % (descr, tmp) )
        return value


    def _GetLimits(self):
        """ Get the limits in world coordinates between which the object exists.
        """
        
        # Obtain untransformed coords (if not an empty set)
        if not self._points:
            return None
        p = self._points.data
        valid = np.isfinite(p[:,0]) * np.isfinite(p[:,1]) * np.isfinite(p[:,2])
        validpoints = p[valid, :]
        x1, y1, z1 = validpoints.min(axis=0)
        x2, y2, z2 = validpoints.max(axis=0)
        
        # There we are
        return Wobject._GetLimits(self, x1, x2, y1, y2, z1, z2)


    ## Create properties


    @PropWithDraw
    def lw():
        """ Get/Set the lineWidth: the width of the line in pixels.
        If zero, the line is not drawn.
        """
        def fget(self):
            return self._lw
        def fset(self, value):
            self._lw = self._AsFloat(value, 'lineWidth')
        return locals()
    
    @PropWithDraw
    def ls():
        """ Get/Set the lineStyle: the style of the line.
          * Solid line: '-'
          * Dotted line: ':'
          * Dashed line: '--'
          * Dash-dot line: '-.' or '.-'
          * A line that is drawn between each pair of points: '+'
          * No line: '' or None.
        """
        def fget(self):
            return self._ls
        def fset(self, value):
            if not value:
                value = None
            elif not isinstance(value, basestring):
                raise Exception("Error in lineStyle: style must be a string!")
            elif value not in ['-', '--', ':', '-.', '.-', '+']:
                raise Exception("Error in lineStyle: unknown line style!")
            self._ls = value
        return locals()
    
    @PropWithDraw
    def lc():
        """ Get/Set the lineColor: the color of the line, as a 3-element
        tuple or as a single character string (shown in uppercase):
        Red, Green, Blue, Yellow, Cyan, Magenta, blacK, White.
        """
        def fget(self):
            return self._lc
        def fset(self, value):
            value = getColor(value, 'lineColor')
            self._lc = value
        return locals()
    

    @PropWithDraw
    def mw():
        """ Get/Set the markerWidth: the width (bounding box) of the marker
        in (screen) pixels. If zero, no marker is drawn.
        """
        def fget(self):
            return self._mw
        def fset(self, value):
            self._mw = self._AsFloat(value, 'markerWidth')
        return locals()
    
    @PropWithDraw
    def ms():
        """ Get/Set the markerStyle: the style of the marker.
          * Plus: '+'
          * Cross: 'x'
          * Square: 's'
          * Diamond: 'd'
          * Triangle (pointing up, down, left, right): '^', 'v', '<', '>'
          * Pentagram star: 'p' or '*'
          * Hexgram: 'h'
          * Point/cirle: 'o' or '.'
          * No marker: '' or None
        """
        def fget(self):
            return self._ms
        def fset(self, value):
            if not value:
                value = None
            elif not isinstance(value, basestring):
                raise Exception("markerstyle (ms) should be a string!")
            elif value not in 'sd+x*phfv^><.o':
                raise Exception("Error in markerStyle: unknown line style!")
            self._ms = value
        return locals()
    
    @PropWithDraw
    def mc():
        """ Get/Set the markerColor: The color of the face of the marker
        If None, '', or False, the marker face is not drawn (but the edge is).
        """
        def fget(self):
            return self._mc
        def fset(self, value):
            self._mc = getColor(value, 'markerColor')
        return locals()
    
    @PropWithDraw
    def mew():
        """ Get/Set the markerEdgeWidth: the width of the edge of the marker.
        If zero, no edge is drawn.
        """
        def fget(self):
            return self._mew
        def fset(self, value):
            self._mew = self._AsFloat(value, 'markerEdgeWidth')
        return locals()
    
    @PropWithDraw
    def mec():
        """ Get/Set the markerEdgeColor: the color of the edge of the marker.
        """
        def fget(self):
            return self._mec
        def fset(self, value):
            self._mec = getColor(value, 'markerEdgeColor')
        return locals()
    
#     # create aliases
#     lineWidth = lw
#     lineStyle = ls
#     lineColor = lc
#     markerWidth = mw
#     markerStyle = ms
#     markerColor = mc
#     markerEdgeWidth = mew
#     markerEdgeColor = mec

    @PropWithDraw
    def alpha():
        """ Get/Set the alpha (transparancy) of the line and markers.
        When this is < 1, the line cannot be anti-aliased, and it
        is drawn on top of any other wobjects.
        """
        def fget(self):
            return self._alpha1
        def fset(self, value):
            self._alpha1 = self._AsFloat(value, 'alpha')
        return locals()
    
    ## Set methods

    @DrawAfter
    def SetXdata(self, data):
        """ SetXdata(data)
        
        Set the x coordinates of the points of the line.
        
        """
        self._points[:,0] = handleInvalidValues(data)
    
    @DrawAfter
    def SetYdata(self, data):
        """ SetYdata(data)
        
        Set the y coordinates of the points of the line.
        
        """
        self._points[:,1] = handleInvalidValues(data)
    
    @DrawAfter
    def SetZdata(self, data):
        """ SetZdata(data)
        
        Set the z coordinates of the points of the line.
        
        """
        self._points[:,2] = handleInvalidValues(data)
    
    @DrawAfter
    def SetPoints(self, points):
        """ SetPoints(points)
        
        Set x,y (and optionally z) data. The given argument can be anything
        that can be converted to a pointset. (From version 1.7 this method
        also works with 2D pointsets.)
        
        The data is copied, so changes to original data will not affect
        the visualized points. If you do want this, use the points property.
        
        """
        
        # Try make it a (copied) pointset (handle masked array)
        if is_Pointset(points):
            points = Pointset(handleInvalidValues(points.data))
        else:
            points = Pointset(handleInvalidValues(points))
       
        # Add z dimension to points if not available
        if points.ndim == 3:
            pass
        elif points.ndim == 2:
            zz = 0.1*np.ones((len(points._data),1), dtype='float32')
            points._data = np.concatenate((points._data, zz),1)
        elif points.ndim == 1:
            N = len(points._data)
            xx = np.arange(N, dtype='float32').reshape(N, 1)
            zz = 0.1*np.ones((N,1), dtype='float32')
            points._data = np.concatenate((xx, points._data, zz), 1)
        
        # Store
        self._points = points
    
    @property
    def points(self):
        """ Get a reference to the internal Pointset used to draw the line
        object. Note that this pointset is always 3D. One can modify
        this pointset in place, but note that a call to Draw() may be
        required to update the screen. (New in version 1.7.)
        """
        return self._points
    

    ## Draw methods

    def OnDrawFast(self):
        self.OnDraw(True)

    def OnDraw(self, fast=False):

        # add z dimension to points if not available
        pp = self._points
        if pp.ndim == 2:
            # a bit dirty this
            tmp = pp._data, 0.1*np.ones((len(pp._data),1),dtype='float32')
            pp._data = np.concatenate(tmp,1)

        # can I draw this data?
        if pp.ndim != 3:
            raise Exception("Can only draw 3D data!")

        # no need to draw if no points
        if len(self._points) == 0:
            return

        # enable anti aliasing and blending
        gl.glEnable(gl.GL_LINE_SMOOTH)
        gl.glEnable(gl.GL_BLEND)

        # lines
        if self.lw and self.ls:
            self._DrawLines()

        # points
        if self.mw and self.ms:
            self._DrawPoints()

        # clean up
        #gl.glDisable(gl.GL_BLEND)


    def _DrawLines(self):

        # set stipple style
        if not self.ls in lineStyles:
            stipple = False
        else:
            stipple = lineStyles[self.ls]
        #
        if stipple and self.lw:
            gl.glEnable(gl.GL_LINE_STIPPLE)
            gl.glLineStipple(int(round(self.lw)), stipple)
        else:
            gl.glDisable(gl.GL_LINE_STIPPLE)

        # init vertex array
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glVertexPointerf(self._points.data)
        
        # linepieces drawn on top of other should draw just fine. See issue #95
        gl.glDepthFunc(gl.GL_LEQUAL)
        
        # init blending. Only use constant blendfactor when alpha<1
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        if self._alpha1<1:
            #if len(self._points) < 1000:
            if getOpenGlCapable('1.4','transparant points and lines'):
                gl.glBlendFunc(gl.GL_CONSTANT_ALPHA, gl.GL_ONE_MINUS_CONSTANT_ALPHA)
                gl.glBlendColor(0.0,0.0,0.0, self._alpha1)
            gl.glDisable(gl.GL_DEPTH_TEST)
        
        # get color
        clr = getColor( self.lc )

        if clr and self._alpha1>0:

            # set width and color
            gl.glLineWidth(self.lw)
            gl.glColor3f(clr[0], clr[1], clr[2])

            # draw
            method = gl.GL_LINE_STRIP
            if self.ls == '+':
                method = gl.GL_LINES
            gl.glDrawArrays(method, 0, len(self._points))
            # flush!
            gl.glFlush()

        # clean up
        gl.glDisable(gl.GL_LINE_STIPPLE)
        gl.glLineStipple(int(round(self.lw)), int('1111111111111111',2))
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
        gl.glDepthFunc(gl.GL_LESS)


    def _DrawPoints(self):

        # get colors (use color from edge or face if not present)
        clr1 = getColor(self.mc)
        clr2 = getColor(self.mec)

        # draw face or edge?
        drawFace = bool(self.mc) # if not ms or mw we would not get here
        drawEdge = self.mec and self.mew
        if not drawFace and not drawEdge:
            return

        # get figure
        f = self.GetFigure()
        if not f:
            return

        # init blending. Only use constant blendfactor when alpha<1
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        if self._alpha1<1:
            if getOpenGlCapable('1.4','transparant points and lines'):
                gl.glBlendFunc(gl.GL_CONSTANT_ALPHA,
                    gl.GL_ONE_MINUS_CONSTANT_ALPHA)
                gl.glBlendColor(0.0,0.0,0.0, self._alpha1)
            gl.glDisable(gl.GL_DEPTH_TEST)

        # init vertex array
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glVertexPointerf(self._points.data)

        # points drawn on top of points should draw (because we draw
        # the face and edge seperately)
        gl.glDepthFunc(gl.GL_LEQUAL)

        # Enable alpha test, such that fragments with 0 alpha
        # will not update the z-buffer.
        gl.glEnable(gl.GL_ALPHA_TEST)
        gl.glAlphaFunc(gl.GL_GREATER, 0.01)

        if self.ms in ['o','.','s'] and not drawEdge:
            # Use standard OpenGL points, faster and anti-aliased
            # Pure filled points or squares always work.

            # choose style
            if self.ms == 's':
                gl.glDisable(gl.GL_POINT_SMOOTH)
            else:
                gl.glEnable(gl.GL_POINT_SMOOTH)

            # draw faces only
            if drawFace:
                gl.glColor3f(clr1[0],clr1[1],clr1[2])
                gl.glPointSize(self.mw)
                gl.glDrawArrays(gl.GL_POINTS, 0, len(self._points))

        elif self.ms in ['o','.','s'] and drawFace and self.alpha==1:
            # Use standard OpenGL points, faster and anti-aliased
            # If alpha=1 and we have a filled marker, we can draw in two steps.

            # choose style
            if self.ms == 's':
                gl.glDisable(gl.GL_POINT_SMOOTH)
            else:
                gl.glEnable(gl.GL_POINT_SMOOTH)

            # draw edges
            if drawEdge:
                gl.glColor3f(clr2[0],clr2[1],clr2[2])
                gl.glPointSize(self.mw+self.mew*2)
                gl.glDrawArrays(gl.GL_POINTS, 0, len(self._points))
            # draw faces
            if drawFace:
                gl.glColor3f(clr1[0],clr1[1],clr1[2])
                gl.glPointSize(self.mw)
                gl.glDrawArrays(gl.GL_POINTS, 0, len(self._points))

        #elif self.alpha>0:
        else:
            # Use sprites
            
            # get sprites
            tmp = f._markerManager.GetSprites(self.ms, self.mw, self.mew)
            pSize, sprite1, sprite2 = tmp
            gl.glPointSize(pSize)
            
            # draw points for the edges
            if drawEdge:
                sprite2.Enable()
                gl.glColor3f(clr2[0],clr2[1],clr2[2])
                gl.glDrawArrays(gl.GL_POINTS, 0, len(self._points))
            # draw points for the faces
            if drawFace:
                sprite1.Enable()
                gl.glColor3f(clr1[0],clr1[1],clr1[2])
                gl.glDrawArrays(gl.GL_POINTS, 0, len(self._points))
            
            # disable sprites
            sprite1.Disable() # Could as well have used sprite2
        
        # clean up
        gl.glDisable(gl.GL_ALPHA_TEST)
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
        gl.glDepthFunc(gl.GL_LESS)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)


    def OnDrawShape(self, clr):
        # Draw the shape of the line so we can detect mouse actions

        # disable anti aliasing and blending
        gl.glDisable(gl.GL_LINE_SMOOTH)
        gl.glDisable(gl.GL_BLEND)

        # no stippling, square points
        gl.glDisable(gl.GL_LINE_STIPPLE)
        gl.glDisable(gl.GL_POINT_SMOOTH)

        # init vertex array
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glVertexPointerf(self._points.data)

        # detect which parts to draw
        drawLine, drawMarker = False, False
        if self.lw and self.ls and getColor(self.lc):
            drawLine = True
        if self.mw and self.ms:
            drawMarker = True

        if drawLine:
            # set width and color
            gl.glLineWidth(self.lw)
            gl.glColor3f(clr[0], clr[1], clr[2])
            # draw
            gl.glDrawArrays(gl.GL_LINE_STRIP, 0, len(self._points))
            gl.glFlush()

        if drawMarker:
            w = self.mw
            if self.mec:
                w += self.mew
            # set width and color
            gl.glColor3f(clr[0],clr[1],clr[2])
            gl.glPointSize(w)
            # draw
            gl.glDrawArrays(gl.GL_POINTS, 0, len(self._points))
            gl.glFlush()

        # clean up
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)


    def OnDestroy(self):
        # clean up some memory
        self._points.clear()


# This is a new type of wobject called PolarLine which encapsulates
# polar plot data.
class PolarLine(Line):
    """ PolarLine(parent, angle(radians), mag)
    
    The Polarline class represents a set of points (locations)
    in world coordinates. This class can draw lines between the points,
    markers at the point coordinates, or both.
    
    There are several linestyles that can be used:
      * -  a solid line
      * :   a dotted line
      * --  a dashed line
      * -.  a dashdot line
      * .-  dito
      * +   draws a line between each pair of points (handy for visualizing
            for example vectore fields)
    If None, '' or False is given no line is drawn.

    There are several marker styles that can be used:
      * `+`  a plus
      * `x`  a cross
      * `s`  a square
      * `d`  a diamond
      * `^v<>` an up-, down-, left- or rightpointing triangle
      * `*` or `p`  a (pentagram star)
      * `h`  a hexagram
      * `o` or `.`  a point/circle
    If None, '', or False is given, no marker is drawn.

    Performance tip
    ---------------
    The s, o (and .) styles can be drawn using standard
    OpenGL points if alpha is 1 or if no markeredge is drawn.
    Otherwise point sprites are used, which can be slower
    on some cards (like ATI, Nvidia performs quite ok with with
    sprites)
    
    """
    def __init__(self, parent, angs, mags):
        self._angs = angs
        self._mags = mags
        x = mags * np.cos(angs)
        y = mags * np.sin(angs)
        z = np.zeros((np.size(x), 1))
        tmp = x, y, z
        pp = Pointset(np.concatenate(tmp, 1))
        Line.__init__(self, parent, pp)

    def TransformPolar(self, radialRange, angRefPos, sense):
        offsetMags = self._mags - radialRange.min
        rangeMags = radialRange.range
        offsetMags[offsetMags > rangeMags] = rangeMags
        tmpang = angRefPos + sense * self._angs
        x = offsetMags * np.cos(tmpang)
        y = offsetMags * np.sin(tmpang)
        z = np.zeros((np.size(x), 1)) + 0.2
        x[offsetMags < 0] = 0
        y[offsetMags < 0] = 0
        tmp = x, y, z
        self._points = Pointset(np.concatenate(tmp, 1))

    def _GetPolarLimits(self):
        if not self._points:
            return None
        else:
            return Range(self._angs.min(), self._angs.max()), \
                   Range(self._mags.min(), self._mags.max())



def handleInvalidValues(values):
    """ handleInvalidValues(values)
    
    Modifies any invalid values (NaN, Inf, -Inf) to Inf,
    and turn masked values of masked arrays to Inf.
    Returns a copy if correction if needed.
    """
    if isinstance(values, np.ma.MaskedArray):
        values = values.filled(np.inf)
        _inplace = True  # values is already a copy, so we can modify it
    else:
        _inplace = False
    if not isinstance(values, np.ndarray):
        values = np.array(values)
    
    invalid = ~np.isfinite(values)
    # Determine if we should make a copy
    if invalid.sum() and not _inplace:
        values = values.copy()
    # Convert values and return
    try:
        values[invalid] = np.inf
    except OverflowError:
        pass # Cannot do this with integer types
    return values



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" The FLTK backend.


"""

import visvis
from visvis import BaseFigure, events, constants
from visvis.core.misc import basestring

import fltk

KEYMAP = {  fltk.FL_SHIFT: constants.KEY_SHIFT,
            fltk.FL_Shift_L: constants.KEY_SHIFT,
            fltk.FL_Shift_R: constants.KEY_SHIFT,
            fltk.FL_ALT : constants.KEY_ALT,
            fltk.FL_Alt_L : constants.KEY_ALT,
            fltk.FL_Alt_R : constants.KEY_ALT,
            fltk.FL_Control_L: constants.KEY_CONTROL,
            fltk.FL_Control_R: constants.KEY_CONTROL,
            fltk.FL_Left: constants.KEY_LEFT,
            fltk.FL_Up: constants.KEY_UP,
            fltk.FL_Right: constants.KEY_RIGHT,
            fltk.FL_Down: constants.KEY_DOWN,
            fltk.FL_Page_Up: constants.KEY_PAGEUP,
            fltk.FL_Page_Down: constants.KEY_PAGEDOWN,
            fltk.FL_Delete: constants.KEY_DELETE
            }

# Make uppercase letters be lowercase
for i in range(ord('A'), ord('Z')):
    KEYMAP[i] = i+32


def modifiers():
    """Convert the fltk modifier state into a tuple of active modifier keys."""
    mod = ()
    fltkmod = fltk.Fl.event_state()
    if fltk.FL_SHIFT & fltkmod:
        mod += constants.KEY_SHIFT,
    if fltk.FL_CTRL & fltkmod:
        mod += constants.KEY_CONTROL,
    if fltk.FL_ALT & fltkmod:
        mod += constants.KEY_ALT,
    return mod


class GLWidget(fltk.Fl_Gl_Window):
    """ Implementation of the GL_window, which passes a number of
    events to the Figure object that wraps it.
    """
    
    def __init__(self, figure, *args, **kwargs):
        fltk.Fl_Gl_Window.__init__(self, *args, **kwargs)
        self.figure = figure
        
        # Callback when closed
        self.callback(self.OnClose)
    
    
    def handle(self, event):
        """ All events come in here. """
        if not self.figure:
            return 1
        
        # map fltk buttons to visvis buttons
        buttons = [0,1,0,2,0,0,0]
        
        if event == fltk.FL_PUSH:
            x, y = fltk.Fl.event_x(), fltk.Fl.event_y()
            but = buttons[fltk.Fl.event_button()]
            self.figure._GenerateMouseEvent('down', x, y, but, modifiers())
        
        elif event == fltk.FL_RELEASE:
            x, y = fltk.Fl.event_x(), fltk.Fl.event_y()
            but = buttons[fltk.Fl.event_button()]
            if fltk.Fl.event_clicks() == 1:
                # double click
                self.figure._GenerateMouseEvent('double', x, y, but, modifiers())
            else:
                # normal release
                self.figure._GenerateMouseEvent('up', x, y, but, modifiers())
        
        elif event in [fltk.FL_MOVE, fltk.FL_DRAG]:
            w, h = self.w(), self.h()  # noqa - not used?
            self.OnMotion(None)
        elif event == fltk.FL_MOUSEWHEEL:
            self.OnMouseWheel(None)
        elif event == fltk.FL_ENTER:
            ev = self.figure.eventEnter
            ev.Set(0,0,0)
            ev.Fire()
        elif event == fltk.FL_LEAVE:
            ev = self.figure.eventLeave
            ev.Set(0,0,0)
            ev.Fire()
        elif event == fltk.FL_KEYDOWN:
            self.OnKeyDown(None)
        elif event == fltk.FL_KEYUP:
            self.OnKeyUp(None)
        elif event == fltk.FL_CLOSE:
            self.OnClose(None)
        
        elif event == fltk.FL_FOCUS:
            self.OnFocus(None)
        else:
            return 1 # maybe someone else knows what to do with it
        return 1 # event was handled.
    
    
    def resize(self, x, y, w, h):
        # Overload resize function to also draw after resizing
        if self.figure:
            fltk.Fl_Gl_Window.resize(self, x, y, w, h)
            self.figure._OnResize()
    
    def draw(self):
        # Do the draw commands now
        self.figure.OnDraw()
    
    
    def OnMotion(self, event):
        # prepare and fire event
        x, y = fltk.Fl.event_x(), fltk.Fl.event_y()
        self.figure._GenerateMouseEvent('motion', x, y, 0, modifiers())
    
    def OnMouseWheel(self, event):
        # Mpff, I cant event test horizontal scrolling, because if I do, fltk segfaults
        horizontal = -1.0 * fltk.Fl.event_dx() or 0
        vertical = -1.0 * fltk.Fl.event_dy() or 0
        x, y = fltk.Fl.event_x(), fltk.Fl.event_y()
        self.figure._GenerateMouseEvent('scroll', x, y, horizontal, vertical, modifiers())
    
    def OnKeyDown(self, event):
        key, text = self._ProcessKey()
        self.figure._GenerateKeyEvent('keydown', key, text, modifiers())
    
    def OnKeyUp(self, event):
        key, text = self._ProcessKey()
        self.figure._GenerateKeyEvent('keyup', key, text, modifiers())
    
    def _ProcessKey(self):
        """ evaluates the keycode of fltk, and transform to visvis key.
        Also produce text version.
        return key, text. """
        key = fltk.Fl.event_key()
        if isinstance(key, basestring):
            key = ord(key)
        # special cases for shift control and alt -> map to 17 18 19
        if key in KEYMAP:
            return KEYMAP[key], ''
        else:
            # other key, try producing text
            #print(key, self._shiftDown)
            if (97 <= key <= 122) and fltk.Fl.event_shift():
                key -= 32
            try:
                return key, chr(key)
            except ValueError:
                return key, ''
    
    
    def OnClose(self, event=None):
        if self.figure:
            self.figure.Destroy()
        self.hide()
    
    
    def OnFocus(self, event):
        BaseFigure._currentNr = self.figure.nr


class Figure(BaseFigure):
    """ This is the fltk implementation of the figure class.
    
    A Figure represents the OpenGl context and is the root
    of the visualization tree; a Figure Wibject does not have a parent.
    
    A Figure can be created with the function vv.figure() or vv.gcf().
    """
    
    def __init__(self, *args, **kwargs):
        
        self._widget = None
        self._widget_args = (args, kwargs)
        if kwargs.get('create_widget', True):
            self.CreateWidget()
        
        # call original init AFTER we created the widget
        BaseFigure.__init__(self)
    
    def CreateWidget(self):
        """ Create the Figure's widget if necessary, and return the
        widget. """
        if self._widget is None:
            # Make sure there is a native app and the timer is started
            # (also when embedded)
            app.Create()
            
            # create widget
            updatePosition = False
            args, kwargs = self._widget_args
            if 'create_widget' in kwargs:
                updatePosition = True
                del(kwargs['create_widget'])
            self._widget = GLWidget(self, *args, **kwargs)
            if updatePosition:
                self.position._Changed()
        return self._widget
    
    def _SetCurrent(self):
        """ make this scene the current context """
        if self._widget:
            self._widget.make_current()
        
        
    def _SwapBuffers(self):
        """ Swap the memory and screen buffer such that
        what we rendered appears on the screen """
        if self._widget:
            self._widget.swap_buffers()

    def _SetTitle(self, title):
        """ Set the title of the figure... """
        if self._widget:
            window = self._widget
            if hasattr(window,'label'):
                window.label(title)
    
    def _SetPosition(self, x, y, w, h):
        """ Set the position of the widget. """
        if self._widget:
            # select widget to resize. If it
            widget = self._widget
            # apply
            widget.position(x,y)
            widget.size(w, h)

    def _GetPosition(self):
        """ Get the position of the widget. """
        if self._widget:
            # select widget to resize. If it
            widget = self._widget
            # get and return
            return widget.x(), widget.y(), widget.w(), widget.h()
        return 0, 0, 0, 0
    
    def _RedrawGui(self):
        if self._widget:
            self._widget.redraw()
    
    def _ProcessGuiEvents(self):
        fltk.Fl.wait(0)
    
    def _Close(self, widget=None):
        if widget is None:
            widget = self._widget
        if widget:
            widget.OnClose()
    

def newFigure():
    """ Create a window with a figure widget.
    """
    
    # Create figure
    size = visvis.settings.figureSize
    figure = Figure(size[0], size[1], "Figure")
    figure._widget.size_range(100,100,0,0,0,0)
    figure._widget.show() # Show AFTER canvas is added
    
    # Make OpenGl Initialize and return
    # Also call draw(), otherwise it will not really draw and crash on Linux
    figure.DrawNow()
    figure._widget.draw()
    return figure




class VisvisEventsTimer:
    """ Timer that can be started and stopped.
    """
    def __init__(self):
        self._running = False
    def Start(self):
        if not self._running:
            self._running = True
            self._PostOneShot()
    def Stop(self):
        self._running = False
    def _PostOneShot(self):
        fltk.Fl.add_timeout(0.01, self._Fire)
    def _Fire(self):
        if self._running:
            events.processVisvisEvents()
            self._PostOneShot() # Repost


class App(events.App):
    """ App()
    
    Application class to wrap the GUI applications in a class
    with a simple interface that is the same for all backends.
    
    This is the fltk implementation.
    
    """
    
    def __init__(self):
        # Init timer
        self._timer = VisvisEventsTimer()
    
    def _GetNativeApp(self):
        self._timer.Start()
        return fltk.Fl
    
    def _ProcessEvents(self):
        app = self._GetNativeApp()
        app.wait(0)
    
    def _Run(self):
        app = self._GetNativeApp()
        if hasattr(app, '_in_event_loop') and app._in_event_loop:
            pass # Already in event loop
        else:
            app.run()

# Create application instance now
app = App()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.


import OpenGL.GL as gl
import OpenGL.GLU as glu

import numpy as np

import visvis as vv
#
from visvis.core.baseTexture import TextureObject
from visvis.core.base import Wobject
from visvis.core.misc import Property, PropWithDraw, basestring, unichr
from visvis.core.misc import getColor
#
from visvis.core.cameras import depthToZ
from visvis.core.baseWibjects import Box


escapes = {
    # upper case greek
    'Alpha':0x0391, 'Beta':0x0392, 'Gamma':0x0393, 'Delta':0x0394,
    'Epsilon':0x0395, 'Zeta':0x0396, 'Eta':0x0397, 'Theta':0x0398,
    'Iota':0x0399, 'Kappa':0x039A, 'Lambda':0x039B, 'Mu':0x039C,
    'Nu':0x039D, 'Xi':0x039E, 'Omicron':0x039F,
    'Pi':0x03A0, 'Rho':0x03A1,             'Sigma':0x03A3, 'Tau':0x03A4,
    'Upsilon':0x03A5, 'Phi':0x03A6, 'Chi':0x03A7, 'Psi':0x03A8, 'Omega':0x03A9,
    # lower case greek
    'alpha':0x03B1, 'beta':0x03B2, 'gamma':0x03B3, 'delta':0x03B4,
    'epsilon':0x03B5, 'zeta':0x03B6, 'eta':0x03B7, 'theta':0x03B8,
    'iota':0x03B9, 'kappa':0x03BA, 'lambda':0x03BB, 'mu':0x03BC,
    'nu':0x03BD, 'xi':0x03BE, 'omicron':0x03BF,
    'pi':0x03C0, 'rho':0x03C1, 'varsigma':0x03C2, 'sigma':0x03C3,
    'tau':0x03C4, 'upsilon':0x03C5,
    'phi':0x03C6, 'chi':0x03C7, 'psi':0x03C8, 'omega':0x03C9,
    # some math
    'Re':0x211c, 'Im':0x2111, 'null':0x2300, 'infty':0x221e,
    'int':0x222b, 'iint':0x222c, 'iiint':0x222d,
    'forall':0x2200,
    'leq':0x2264, 'geq':0x2265, 'approx':0x2248, 'approxeq':0x2243, 'ne':0x2260,
    'in':0x2208,
    'leftarrow':0x2190,'uparrow':0x2191,'rightarrow':0x2192,'downarrow':0x2193,
    'Leftarrow':0x21D0,'Uparrow':0x21D1,'Rightarrow':0x21D2,'Downarrow':0x21D3,
    'leftceil':0x2308,'rightceil':0x2309,'leftfloor':0x230A,'rightfloor':0x230B,
    'times':0x00d7, 'cdot':0x2219, 'pm':0x00b1,
    'oplus':0x2295, 'ominus':0x2296, 'otimes':0x2297, 'oslash':0x2298,
    }

# sort the keys, such that longer names are replaced first
escapesKeys = list(escapes.keys())
escapesKeys.sort( key=len, reverse=True)



class AtlasTexture(TextureObject):
    """ Class to represent a texture that contains characters.
    """
    
    def __init__(self):
        TextureObject.__init__(self, 2)
    
    def _UploadTexture(self, data, *args):
        """ Overload to make it an alpha map.
        """
        
        # Add lumincance channel
        # data2 = np.zeros((data.shape[0],data.shape[1],2), dtype=np.uint8)
        # data2[:,:,0] = 255
        # data2[:,:,1] = data
        
        shape = data.shape
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, 2, shape[1],shape[0], 0,
            gl.GL_ALPHA, gl.GL_UNSIGNED_BYTE, data)  # See issue #110
            # gl.GL_LUMINANCE_ALPHA, gl.GL_UNSIGNED_BYTE, data2)
        
        tmp1, tmp2 = gl.GL_LINEAR, gl.GL_LINEAR
        #tmp1, tmp2 = gl.GL_NEAREST, gl.GL_NEAREST
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, tmp1)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, tmp2)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP)

    
    def _UpdateTexture(self, data, *args):
        """ Overload to make it an alpha map.
        """
        # Add lumincance channel
        # data2 = np.zeros((data.shape[0],data.shape[1],2), dtype=np.uint8)
        # data2[:,:,0] = 255
        # data2[:,:,1] = data
        
        shape = data.shape
        gl.glTexSubImage2D(gl.GL_TEXTURE_2D, 0, 0,0, shape[1],shape[0],
            gl.GL_ALPHA, gl.GL_UNSIGNED_BYTE, data)
            # gl.GL_LUMINANCE_ALPHA, gl.GL_UNSIGNED_BYTE, data2)


class FontManager(object):
    """ The font manager is a class that can generate text.
    There is one fontmanager per context (i.e. Figure).
    """
    
    # To create a font manager, subclass this class and implement the
    # Compile, Position and Draw methods that each have a textObject
    # (Text or Label instance) as the first argument.
    #
    # * In Compile() the text should be compiled. This function is called
    #   as little as possible; do most of the heavy lifting here.
    #   Use BaseText._SetCompiledData() to buffer the result.
    #
    # * In Position() one can implement the final positioning of the text.
    #   Use BaseText._GetCompiledData() to get the buffered data and use
    #   BaseText._SetFinalData() to buffer the result.
    #
    # * In Draw() the text should be drawn to the screen. Use
    #   BaseText._GetFinalData() to get the data for drawing.
    #
    # The _SetXxxData and _GetXxxData may contain any values.
    # The only restriction is that the first argument is the array of vertices.
    
    def ConvertEscapedText(self, tt):
        tt = tt.replace(r'\\', unichr(0))
        # Catch valid escape codes that were supposed to be greek letters
        # (a b f n r t v)
        tt = tt.replace('\alpha', unichr(escapes['alpha']))
        tt = tt.replace('\beta', unichr(escapes['beta']))
        tt = tt.replace('\null', unichr(escapes['null']))
        tt = tt.replace('\nu', unichr(escapes['nu']))
        tt = tt.replace('\rho', unichr(escapes['rho']))
        tt = tt.replace('\theta', unichr(escapes['theta']))
        tt = tt.replace('\tau', unichr(escapes['tau']))
        tt = tt.replace('\forall', unichr(escapes['forall']))
        tt = tt.replace('\approxeq', unichr(escapes['approxeq']))
        tt = tt.replace('\approx', unichr(escapes['approx']))
        tt = tt.replace('\ne', unichr(escapes['ne']))
        tt = tt.replace('\rightarrow', unichr(escapes['rightarrow']))
        tt = tt.replace('\rightceil', unichr(escapes['rightceil']))
        tt = tt.replace('\rightfloor', unichr(escapes['rightfloor']))
        tt = tt.replace('\times', unichr(escapes['times']))
        for c in escapesKeys:
            tt = tt.replace('\\'+c, unichr(escapes[c]))
        # get italic and bold modifiers
        tt = tt.replace('\i', '\x06') # just use some char that is no string
        tt = tt.replace('\b', '\x07') # Note that '\i' == r'\i', but '\b' != r'\b'
        tt = tt.replace(r'\b', '\x07')
        tt = tt.replace(unichr(0), r'\\')
        return tt
    
    
    def GetFont(self, fontName, fontSize, bold=False, italic=False):
        """ Returns something that Compile() can work with.
        """
        raise NotImplementedError()
    
    def Compile(self, textObject):
        pass
    
    def Position(self, textObject):
        if textObject._compiledData is None:
            self.Compile(textObject)
        
    def Draw(self, textObject, x=0, y=0, z=0):
        if textObject._finalData is None:
            self.Position(textObject)

    
def correctVertices( textObject, vertices, charHeigh):
    """ Provides a default algorithm that can be used
    in the FontManager.Position() method to correct the vertices
    for alignment and angle.
    """
    
    # obtain dimensions
    if len(vertices):
        x1, x2 = vertices[:,0].min(), vertices[:,0].max()
    else:
        x1, x2 = 0,0
    y1, y2 = 0, charHeigh
    
    # set anchor
    if textObject.halign < 0:  anchorx = x1
    elif textObject.halign > 0:  anchorx = x2
    else: anchorx = x1 + (x2-x1)/2.0
    #
    if textObject.valign < 0:  anchory = y1
    elif textObject.valign > 0:  anchory = y2
    else: anchory = y1 + (y2-y1)/2.0
    
    # apply anchor
    angle = textObject.textAngle
    if isinstance(textObject, Text):
        # Text is a wobject, so must be flipped on y axis
        vertices[:,0] = vertices[:,0] - anchorx
        vertices[:,1] = -(vertices[:,1] - anchory)
    
    elif isinstance(textObject, Label):
        angle = -textObject.textAngle
        vertices[:,0] = vertices[:,0] - anchorx
        vertices[:,1] = vertices[:,1] - anchory
    
    # apply angle
    if angle != 0.0:
        cos_angle = np.cos(angle*np.pi/180.0)
        sin_angle = np.sin(angle*np.pi/180.0)
        vertices[:,0], vertices[:,1] = (
            vertices[:,0] * cos_angle - vertices[:,1] * sin_angle,
            vertices[:,0] * sin_angle + vertices[:,1] * cos_angle)
    
    # Move anchor in label
    if isinstance(textObject, Label):
        w,h = textObject.position.size
        # determine whether the text is vertical or horizontal
        halign, valign = textObject.halign, textObject.valign
        if textObject.textAngle > 135 or textObject.textAngle < -135:
            halign, valign = -halign, valign
        elif textObject.textAngle > 45:
            halign, valign = valign, -halign
        elif textObject.textAngle < -45:
            halign, valign = valign, halign
        # set anchor y
        if valign < 0:  anchory = 0
        elif valign > 0:  anchory = h
        else:  anchory = h/2.0
        # set anchor x
        if halign < 0:  anchorx = 0
        elif halign > 0:  anchorx = w
        else:  anchorx = w/2.0
        # apply
        vertices[:,0] = vertices[:,0] + anchorx
        vertices[:,1] = vertices[:,1] + anchory


def simpleTextureDraw(vertices, texcords, texture, color):
    """ Simply draw characters, given vertices, texcords, a texture
    and a color.
    """
    # Make arrays
    if isinstance(vertices, vv.Pointset):
        vertices = vertices.data
    if isinstance(texcords, vv.Pointset):
        texcords = texcords.data
    
    # Enable texture
    texture.Enable()
    
    # Allow overlapping quads to overwrite the empty part of another.
    gl.glDepthFunc(gl.GL_LEQUAL)
    
    # init vertex and texture array
    gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
    gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)
    gl.glVertexPointerf(vertices)
    gl.glTexCoordPointerf(texcords)
    
    # draw
    if color and len(vertices):
        gl.glColor(color[0], color[1], color[2])
        gl.glDrawArrays(gl.GL_QUADS, 0, len(vertices))
        gl.glFlush()
    
    # disable texture and clean up
    texture.Disable()
    gl.glDepthFunc(gl.GL_LESS)
    gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
    gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY)



class BaseText(object):
    """ BaseText(text='', fontName=None, fontSize=9, color='k')
    
    Base object for the Text wobject and Label wibject.
    fontname may be 'mono', 'sans', 'serif' or None, in which case
    the vv.settings.defaultFontName is used.
    """
    
    # The idea is that the FontManager produces the _compiledData
    # for this object in its Compile method. In its Position method it
    # will create _finalData
    
    def __init__(self, text='', fontName=None, fontSize=9, color='k'):
        
        # Check fontname
        if fontName is None:
            fontName = vv.settings.defaultFontName
        
        # Store arguments
        self._text = text
        self._fontName = fontName.lower()
        self._fontSize = fontSize
        self._color = None
        self.textColor = color
        
        # Set more parameters
        self._halign = -1
        self._valign = 0
        self._angle = 0
        
        # Initialize some attributes
        self.Invalidate()
        
        # Try getting the font manager now
        self._lastFontManager = None
        self._GetFontManager()
    
    
    def _GetFontManager(self):
        """ Convienence function to obtain a reference to the font manager.
        The font manager is stored at the figure; there is one font
        manager per OpenGl contex.
        """
        fig = self.GetFigure()
        if fig:
            self._lastFontManager = fig._fontManager
        return self._lastFontManager
    
    
    def Invalidate(self):
        """ Invalidate this object, such that the text is recompiled
        the next time it is drawn.
        """
        
        # Data for drawing
        self._compiledData = None    # The preliminary (relative) vertex coordinates etc
        self._finalData = None      # The final (absolute) vertex coordinates etc
        
        # Buffered values to store the limits of the vertices
        self._deltax = 0, 0
        self._deltay = 0, 0
    
    def _SetCompiledData(self, *args):
        self._compiledData = args
    
    def _GetCompiledData(self):
        return self._compiledData
    
    def _SetFinalData(self, *args):
        self._finalData = args
        
        vertices = args[0]
        if vertices is not None and len(vertices):
            self._deltax = vertices[:,0].min(), vertices[:,0].max()
            self._deltay = vertices[:,1].min(), vertices[:,1].max()
    
    def _GetFinalData(self):
        return self._finalData
    
    
    def GetVertexLimits(self):
        """ Get the limits of the vertex data.
        Returns (xmin, xmax), (ymin, ymax)
        """
        # Make sure we have vertices
        if self._finalData is None:
            self.UpdatePosition()
        return self._deltax, self._deltay
        
    
    def UpdatePosition(self, *args):
        """ Updates the position now, Compiles the text if necessary.
        """
        fm = self._GetFontManager()
        if fm:
            fm.Position(self) # Calls Compile() if necessary
    
    @property
    def isCompiled(self):
        return self._compiledData is not None and len(self._compiledData[0])
    
    @property
    def isPositioned(self):
        return self._finalData is not None and len(self._finalData[0])
    
    
    @Property # Smart draw
    def text():
        """Get/Set the text to display.
        """
        def fget(self):
            return self._text
        def fset(self, value):
            if value != self._text:
                self._text = value
                self.Invalidate() # force recalculation
                self.Draw()
        return locals()
    
    
    @Property  # Smart draw
    def textAngle():
        """Get/Set the angle of the text in degrees.
        """
        def fget(self):
            return self._angle
        def fset(self, value):
            if value != self._angle:
                self._angle = value
                self._finalData = None # force recalculation
                self.Draw()
        return locals()
    
    
    # todo: remove in future version
    @Property
    def textSpacing():
        """Get/Set the spacing between characters.
        """
        def fget(self):
            print('textSpacing property is deprecated.')
            return 1
        def fset(self, value):
            print('textSpacing property is deprecated.')
        return locals()
    
    
    @Property
    def fontSize():
        """Get/Set the size of the text.
        """
        def fget(self):
            return self._fontSize
        def fset(self, value):
            if value != self._fontSize:
                self._fontSize = value
                self.Invalidate() # force recalculation
                self.Draw()
        return locals()
    
    
    @Property
    def fontName():
        """Get/Set the font type by its name.
        """
        def fget(self):
            return self._fontName
        def fset(self, value):
            if value != self._fontName:
                self._fontName = value
                self.Invalidate() # force recalculation
                self.Draw()
        return locals()
    
    
    @Property
    def textColor():
        """Get/Set the color of the text.
        """
        def fget(self):
            return self._color
        def fset(self, value):
            value = getColor(value,'setting textColor')
            if value != self._color:
                self._color = value
                self.Draw()
        return locals()
    
    
    @Property
    def halign():
        """Get/Set the horizontal alignment. Specify as:
          * 'left', 'center', 'right'
          * -1, 0, 1
        """
        def fget(self):
            return self._halign
        def fset(self, value):
            if isinstance(value, int):
                pass
            elif isinstance(value, basestring):
                value = value.lower()
                tmp = {'left':-1,'center':0,'centre':0,'right':1 }
                if not value in tmp:
                    raise ValueError('Invalid value for halign.')
                value = tmp[value.lower()]
            else:
                raise ValueError('halign must be an int or string.')
            value = int(value>0) - int(value<0)
            if value != self._halign:
                self._halign = value
                self._finalData = None # force recalculation
                self.Draw()
        return locals()
    
    
    @Property
    def valign():
        """Get/Set the vertical alignment. Specify as:
          * 'up', 'center', 'down'
          * 'top', 'center', 'bottom'
          * -1, 0, 1
        """
        def fget(self):
            return self._valign
        def fset(self, value):
            if isinstance(value, int):
                pass
            elif isinstance(value, basestring):
                value = value.lower()
                tmp={'up':-1,'top':-1,'center':0,'centre':0,'down':1,'bottom':1}
                if not value in tmp:
                    raise ValueError('Invalid value for valign.')
                value = tmp[value.lower()]
            else:
                raise ValueError('valign must be an int or string.')
            value = int(value>0) - int(value<0)
            if value != self._valign:
                self._valign = value
                self._finalData = None # force recalculation
                self.Draw()
        return locals()



# Label and Text class must be defined here because we need to be able
# to do isinstance() in some methods of Font Managers.

class Text(Wobject, BaseText):
    """ Text(parent, text='', x=0, y=0, z=0, fontName=None, fontSize=9, color='k')
    
    A wobject representing a string of characters. The text has
    a certain position in the scene. The fontname can be
    'mono', 'sans' or 'serif'. If not given, the vv.settings.defaultFontName
    is used.
    
    """
    
    def __init__(self, parent, text='', x=0, y=0, z=0,
                            fontName=None, fontSize=9, color='k'):
        Wobject.__init__(self, parent)
        BaseText.__init__(self, text, fontName, fontSize, color)
        
        # store coordinates
        self._x, self._y, self._z = x, y, z
        
        # for internal use
        self._screenx, self._screeny, self._screenz = 0, 0, 0
    
    
    @PropWithDraw
    def x():
        """Get/Set the x position of the text.
        """
        def fget(self):
            return self._x
        def fset(self, value):
            self._x = value
        return locals()
    
    @PropWithDraw
    def y():
        """Get/Set the y position of the text.
        """
        def fget(self):
            return self._y
        def fset(self, value):
            self._y = value
        return locals()
    
    @PropWithDraw
    def z():
        """Get/Set the z position of the text.
        """
        def fget(self):
            return self._z
        def fset(self, value):
            self._z = value
        return locals()
    
    
    def OnDraw(self):
        # get screen position and store
        tmp = glu.gluProject(self._x, self._y, self._z)
        self._screenx, self._screeny, self._screenz = tuple(tmp)
#         # make integer (to prevent glitchy behaviour), but not z!
#         self._screenx = int(self._screenx+0.5)
#         self._screeny = int(self._screeny+0.5)
    
    
    def OnDrawScreen(self):
        fm = self._GetFontManager()
        if fm:
            fm.Draw(self,
                    self._screenx, self._screeny, depthToZ(self._screenz) )



class Label(Box, BaseText):
    """ Label(parent, text='', fontName=None, fontSize=9, color='k')
    
    A wibject (inherits from box) with text inside.
    The fontname can be 'mono', 'sans' or 'serif'. If not given, the
    vv.settings.defaultFontName is used.
    
    """
    
    def __init__(self, parent, text='', fontName=None, fontSize=9, color='k'):
        Box.__init__(self, parent)
        BaseText.__init__(self, text, fontName, fontSize, color)
        
        # no edge
        self.edgeWidth = 0
        
        # init position (this is to set the size)
        self.position = 10,10,100,16
        
        # we need to know about position changes to update alignment
        self.eventPosition.Bind(self.UpdatePosition)
    
    
    def OnDraw(self):
        
        # Draw the box
        Box.OnDraw(self)
        
        # Draw the text
        fm = self._GetFontManager()
        if fm:
            fm.Draw(self)



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# SSDF is distributed under the terms of the (new) BSD License.
# See http://www.opensource.org/licenses/bsd-license.php

""" ssdf.classmaneger.py

Implements the code to register classes at ssdf for automatic
conversion (kinda like pickling).

"""

import sys
from . import __version__

class _ClassManager:
    """ _ClassManager
    
    This static class enables registering classes, which
    can then be stored to ssdf and loaded from ssdf.
    
    On module loading, the sys module is given a reference to the
    ClassManager called '_ssdf_class_manager'. Other classes can
    thereby register themselves without knowing how to import
    ssdf (provided that ssdf is already imported).
    
    """
    
    # Global dict with registered classes
    _registered_classes = {}
    
    @classmethod
    def _register_at_sys(manager):
        """ _register_at_sys()
        
        Register the manager at the sys module if there is not
        already one of a higher version.
        
        """
        if hasattr(sys, '_ssdf_class_manager'):
            other = sys._ssdf_class_manager
            if manager.__version__() >= other.__version__():
                sys._ssdf_class_manager = manager
                manager._registered_classes.update(other._registered_classes)
                return manager
            else:
                return other
        else:
            sys._ssdf_class_manager = manager
            return manager
    
    @classmethod
    def __version__(manager):
        return __version__
    
    
    @classmethod
    def is_compatible_class(manager, cls):
        """ is_compatible_class(cls)
        
        Returns True if the given class is SSDF-compatible.
        
        """
        return not manager.is_incompatible_class(cls)
    
    
    @classmethod
    def is_incompatible_class(manager, cls):
        """ is_incompatible_class(cls)
        
        Returns a string giving the reason why the given class
        if not SSDF-compatible. If the class is compatible, this
        function returns None.
        
        """
        if not hasattr(cls, '__to_ssdf__'):
            return "class does not have '__to_ssdf__' method"
        if not hasattr(cls, '__from_ssdf__'):
            return "class does not have '__from_ssdf__' classmethod"
        if not isinstance(cls, type):
            return "class is not a type (does not inherit object on Python 2.x)"
    
    
    @classmethod
    def register_class(manager, *args):
        """ register_class(class1, class2, class3, ...)
        
        Register one or more classes. Registered classes can be saved and
        restored from ssdf.
        
        A class needs to implement two methods to qualify for registration:
        * A method __to_ssdf__() that returns an ssdf.Struct
        * A classmethod __from_ssdf__(s) that accepts an ssdf.Struct and
          creates an instance of that class.
        
        """
        for cls in args:
            incomp = manager.is_incompatible_class(cls)
            if incomp:
                raise ValueError('Cannot register class %s: %s.' %
                                                    (cls.__name__, incomp) )
            else:
                manager._registered_classes[cls.__name__] = cls
    
    
    @classmethod
    def is_registered_class(manager, cls):
        """ is_registered_class(cls)
        
        Returns True if the given class is registered.
        
        """
        return cls in manager._registered_classes.values()

# Put in this module namespace and in sys module namespace
# The ClassManager is the latest version
ClassManager = _ClassManager._register_at_sys()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv

def zlabel(text, axes=None):
    """ zlabel(text, axes=None)
    
    Set the zlabel of an axes.
    Note that you can also use "axes.axis.zLabel = text".
    
    Parameters
    ----------
    text : string
        The text to display.
    axes : Axes instance
        Display the image in this axes, or the current axes if not given.
    
    """
    if axes is None:
        axes = vv.gca()
    axes.axis.zLabel = text

if __name__=='__main__':
    # Create a 3D plot
    a = vv.gca()
    vv.plot([1,2,3],[1,3,2],[3,1,2])
    a.cameraType = '3d'
    # Set labels for all 3 dimensions
    vv.xlabel('label x')
    vv.ylabel('label y')
    vv.zlabel('label z')
